from __future__ import annotations
from abc import ABC, abstractmethod
from contextvars import ContextVar
from enum import Enum
from functools import total_ordering
from .halide_ import *
from .halide_ import _unique_name, _UnspecifiedType
from inspect import isclass
from typing import Any, Optional
import builtins
import re
import sys
import warnings

def _fail(msg: str):
    raise HalideError(msg)


def _check(cond: bool, msg: str):
    if not cond:
        _fail(msg)

# Basically, a valid C identifier, except:
#
# -- initial _ is forbidden (rather than merely "reserved")
# -- two underscores in a row is also forbidden
_NAME_RE = re.compile(r"^(?!.*__)[A-Za-z0-9][A-Za-z0-9_]*$")

def _is_valid_name(name: str) -> bool:
    if not name:
        return False
    # We forbid this to avoid ambiguity in arguments to call()
    if name == "generator_params":
        return False

    return _NAME_RE.search(name)


def _check_valid_name(name: str) -> str:
    _check(
        _is_valid_name(name),
        "The name '%s' is not valid for a GeneratorParam, Input, or Output." % name,
    )
    return name


# Transmute 'None' into our internal "UnspecifiedType" as a placeholder;
# also add some syntactic sugar, to allow users to alias
# bool -> UInt(1), int -> Int(32), float -> Float(32)
def _sanitize_type(t: object) -> Type:
    if t is None or t is type(None):
        return _UnspecifiedType()
    elif t is bool:
        return UInt(1)
    elif t is int:
        return Int(32)
    elif t is float:
        return Float(32)
    else:
        _check(isinstance(t, Type), "Expected a Halide Type, but saw: %s" % t)
        return t

def _normalize_type_list(types: object) -> list[Type]:
    # Always treat _UnspecifiedType as a non-type
    if types is None:
        types = []
    elif isinstance(types, Type) and types == _UnspecifiedType():
        types = []
    if type(types) is not list:
        types = [types];
    types = [_sanitize_type(t) for t in types]
    return types


_type_string_map = {
    # "bfloat16": BFloat(16),
    "bool": Bool(),
    "float16": Float(16),
    "float32": Float(32),
    "float64": Float(64),
    "int16": Int(16),
    "int32": Int(32),
    "int8": Int(8),
    "uint16": UInt(16),
    "uint32": UInt(32),
    "uint8": UInt(8),
}


def _parse_halide_type(s: str) -> Type:
    _check(s in _type_string_map, "The value %s cannot be parsed as a Type." % s)
    return _type_string_map[s]


def _parse_halide_type_list(s: str) -> list[Type]:
    return [_parse_halide_type(t) for t in s.split(",")]


class Requirement:
    # Name
    _name: str = ""

    # List of the required types, if any. An empty list means
    # no constraints. The list will usually be a single type,
    # except for Outputs that have Tuple-valued results, which
    # can have multiple types.
    _types: list[Type] = []

    # Required dimensions. 0 = scalar. -1 = unconstrained.
    _dimensions: int = -1

    def __init__(self, name: str, types: object, dimensions: int):
        self._name = _check_valid_name(name)
        self._types = _normalize_type_list(types)
        self._dimensions = dimensions
        _check(
            len(self._types) > 0,
            "No type has been specified for %s. Try specifying one in code, or by setting '%s.type' as a GeneratorParam."
            % (self._name, self._name),
        )
        _check(
            self._dimensions >= 0,
            "No dimensionality has been specified for %s. Try specifying one in code, or by setting '%s.dim' as a GeneratorParam."
            % (self._name, self._name),
        )
        self._check_types_and_dimensions(types, dimensions)

    def _check_types_and_dimensions(self, types: list[Type], dimensions: int):
        types = _normalize_type_list(types)
        assert len(types) > 0
        assert dimensions >= 0

        _check(
            len(types) == len(self._types),
            "Type mismatch for %s: expected %d types but saw %d" % (self._name, len(self._types), len(types)),
        )
        for i in range(0, len(types)):
            _check(
                self._types[i] == types[i],
                "Type mismatch for %s:%d: expected %s saw %s" % (self._name, i, self._types[i], types[i]),
            )
        _check(
            dimensions == self._dimensions,
            "Dimensions mismatch for %s: expected %d saw %d" % (self._name, self._dimensions, dimensions),
        )


# GeneratorParam is a string, int, float, bool, Type
class GeneratorParam:
    # This tells PyType that we dynamically set attributes
    # on this object. (We don't, actually, but it defeats
    # PyType's complaints about using len(), arithmetic operators, etc.
    # on GeneratorParam fields, since they are replaced at runtime.)
    _HAS_DYNAMIC_ATTRIBUTES:bool = True

    _name: str = ""
    _value: Any = None

    def __init__(self, value: Any):
        # Don't parse it here, since we don't have our name yet
        self._value = value

    @staticmethod
    def _parse_value(name: str, gp_type: type, value: Any) -> object:
        if gp_type is str:
            return str(value)
        elif gp_type is bool:
            return bool(value)
        elif gp_type is int:
            return int(value)
        elif gp_type is float:
            return float(value)
        elif gp_type is Type:
            if isinstance(value, (list, Type)):
                value = _normalize_type_list(value)
            else:
                value = _parse_halide_type_list(str(value))
            return value
        else:
            _fail("GeneratorParam %s has unsupported type %s" % (name, gp_type))
            return None

    def _make_replacement(self, value: Any, r: Requirement) -> Any:
        assert _is_valid_name(r._name)

        # Check the default value for validity
        GeneratorParam._parse_value(r._name, type(self._value), self._value)

        if value is not None:
            # parse replacement value for validity, and also for type match with default.
            return GeneratorParam._parse_value(r._name, type(self._value), value)
        else:
            return self._value

    def _get_types_and_dimensions(self) -> (list[Type], int):
        # Use dummy type-and-dimensions here so that the ctor won't fail due to undefined status
        return [Int(32)], 0


class InputBuffer(ImageParam):

    def __init__(self, type: Optional[Type], dimensions: Optional[int]):
        type = _sanitize_type(type)
        if dimensions is None:
            dimensions = -1
        super().__init__(type, dimensions, _unique_name())

    def _get_types_and_dimensions(self) -> (list[Type], int):
        return _normalize_type_list(self.type()), self.dimensions()

    def _get_direction_and_kind(self) -> (ArgInfoDirection, ArgInfoKind):
        return ArgInfoDirection.Input, ArgInfoKind.Buffer

    def _make_replacement(self, value: Any, r: Requirement) -> ImageParam:
        assert _is_valid_name(r._name) and len(r._types) == 1 and r._dimensions >= 0
        if value is None:
            return ImageParam(r._types[0], r._dimensions, r._name)

        _check(
            isinstance(value, (Buffer, ImageParam)),
            "Input %s requires an ImageParam or Buffer argument when using call(), but saw %s" % (r._name, str(value)),
        )
        _check(
            value.defined(),
            "Cannot set the value for %s to an undefined value." % r._name,
        )
        if isinstance(value, Buffer):
            im = ImageParam(value.type(), value.dimensions(), r._name)
            im.set(value)
            value = im
        r._check_types_and_dimensions([value.type()], value.dimensions())
        return value


class InputScalar(Param):

    def __init__(self, type: Optional[Type]):
        type = _sanitize_type(type)
        super().__init__(type, _unique_name())

    def _get_types_and_dimensions(self) -> (list[Type], int):
        return _normalize_type_list(self.type()), 0

    def _get_direction_and_kind(self) -> (ArgInfoDirection, ArgInfoKind):
        return ArgInfoDirection.Input, ArgInfoKind.Scalar

    def _make_replacement(self, value: Any, r: Requirement) -> Param:
        assert _is_valid_name(r._name) and len(r._types) == 1 and r._dimensions == 0
        if value is None:
            return Param(r._types[0], r._name)

        if isinstance(value, (int, float, bool)):
            p = Param(r._types[0], r._name)
            p.set(value)
            return p
        else:
            _check(
                isinstance(value, Param),
                "Input %s requires a Param (or scalar literal) argument when using call(), but saw %s." %
                (r._name, str(value)),
            )
            _check(
                value.defined(),
                "Cannot set the value for %s to an undefined value." % r._name,
            )
            r._check_types_and_dimensions([value.type()], 0)
            return value


class OutputBuffer(Func):

    def __init__(self, types: Optional[Type], dimensions: Optional[int]):
        types = _normalize_type_list(types)
        if dimensions is None:
            dimensions = -1
        super().__init__(types, dimensions, _unique_name())
        self._types = types
        self._dimensions = dimensions

    def _get_types_and_dimensions(self) -> (list[Type], int):
        return self._types, self._dimensions

    def _get_direction_and_kind(self) -> (ArgInfoDirection, ArgInfoKind):
        return ArgInfoDirection.Output, ArgInfoKind.Buffer

    def _make_replacement(self, value: Any, r: Requirement) -> Func:
        assert _is_valid_name(r._name) and len(r._types) > 0 and r._dimensions >= 0
        if value is None:
            return Func(r._types, r._dimensions, r._name)

        _check(
            isinstance(value, (Buffer, Func)),
            "Output %s requires a Func or Buffer argument when using call(), but saw %s" % (r._name, str(value)),
        )
        _check(
            value.defined(),
            "Cannot set the value for %s to an undefined value." % r._name,
        )

        value_types = [value.type()] if isinstance(value, Buffer) else value.output_types
        r._check_types_and_dimensions(value_types, value.dimensions())

        if isinstance(value, Buffer):
            # Allow assignment from a Buffer<> to an Output<Buffer<>>;
            # this allows us to use a statically-compiled buffer inside a Generator
            # to assign to an output.
            f = Func(r._types, r._dimensions, r._name)
            f[_] = value[_]
            value = f

        return value


# OutputScalar is just like OutputBuffer, except it is always dimensions = 0
class OutputScalar(OutputBuffer):

    def __init__(self, types: Optional[Type]):
        super().__init__(types, 0)

    def _make_replacement(self, value: Any, r: Requirement) -> Func:
        assert _is_valid_name(r._name) and len(r._types) > 0 and r._dimensions == 0
        if value is None:
            return Func(r._types, 0, r._name)

        _check(
            isinstance(value, Expr),
            "Output %s requires an Expr argument when using call(), but saw %s" % (r._name, str(value)),
        )
        _check(
            value.defined(),
            "Cannot set the value for %s to an undefined value." % r._name,
        )
        r._check_types_and_dimensions(value.type(), 0)
        f = Func(value.type(), 0, r._name)
        f[_] = value
        return f


@total_ordering
class _Stage(Enum):
    generator_created = 0
    gp_created = 1
    gp_replaced = 2
    io_created = 3
    configure_called = 4
    io_replaced = 5
    pipeline_built = 6

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


def _unsorted_cls_dir(cls):
    for base in cls.__bases__:
        yield from _unsorted_cls_dir(base)

    # Don't use dir(): it will alphabetize the result. It's critical that
    # we produce this list in order of declaration, which __dict__ should guarantee
    # in Python 3.7+.
    for k, v in cls.__dict__.items():
        yield (k, v)


_halide_generator_context = ContextVar('halide_generator_context', default=None)


def _generatorcontext_enter(self: GeneratorContext) -> GeneratorContext:
    if not hasattr(self, "_tokens"):
        self._tokens = []
    self._tokens.append(_halide_generator_context.set(self))
    return self


def _generatorcontext_exit(self: GeneratorContext) -> None:
    _halide_generator_context.reset(self._tokens.pop())


class Generator(ABC):
    """Base class for Halide Generators in Python"""

    def context(self):
        return GeneratorContext(self._target, self._autoscheduler)

    def target(self):
        return self._target

    def autoscheduler(self):
        return self._autoscheduler

    def using_autoscheduler(self):
        return bool(self._autoscheduler.name)

    def natural_vector_size(self, type: Type) -> int:
        return self.target().natural_vector_size(type)

    def add_requirement(self, condition: Expr, *args) -> None:
        assert self._stage < _Stage.pipeline_built
        self._pipeline_requirements.append((condition, [*args]))

    @classmethod
    def call(cls, *args, **kwargs):
        generator = cls()

        # First, fill in all the GeneratorParams
        # (in case some are tied to Inputs).
        gp = kwargs.pop("generator_params", {})
        _check(isinstance(gp, dict), "generator_params must be a dict")
        for k, v in gp.items():
            generator._set_generatorparam_value(k, v)

        generator._advance_to_stage(_Stage.configure_called)

        _check(
            len(args) <= len(generator._arginfos_in),
            "Generator '%s' allows at most %d positional args, but %d were specified." %
            (generator._get_name(), len(generator._arginfos_in), len(args)))

        inputs_set = []
        for i in range(0, len(args)):
            a = generator._arginfos_in[i]
            k = a.name
            v = args[i]
            _check(k not in inputs_set, "Input %s was specified multiple times." % k)
            inputs_set.append(k)
            generator._bind_input(k, [v])

        input_names = [a.name for a in generator._arginfos_in]
        for k, v in kwargs.items():
            _check(k in input_names, "Unknown input '%s' specified via keyword argument." % k)
            _check(k not in inputs_set, "Input %s specified multiple times." % k)
            inputs_set.append(k)
            generator._bind_input(k, [v])

        _check(
            len(inputs_set) == len(generator._arginfos_in), "Generator '%s' requires %d args, but %d were specified." %
            (generator._get_name(), len(generator._arginfos_in), len(inputs_set)))

        generator._build_pipeline()

        outputs = []
        for o in generator._arginfos_out:
            outputs.append(generator._get_output_func(o.name))

        return outputs[0] if len(outputs) == 1 else tuple(outputs)

    def compile_to_callable(self):
        pipeline = self._build_pipeline()
        arguments = [
            self._get_input_parameter(a.name)._to_argument()
            for a in self._get_arginfos()
            if a.dir == ArgInfoDirection.Input
        ]
        return pipeline.compile_to_callable(arguments, self._target)

    # Make it hard for the user to overwrite any members that are GeneratorParams, Inputs, or Outputs
    def __setattr__(self, name, value):
        r = getattr(self, "_requirements", None)
        s = getattr(self, "_stage", None)
        if r and s and (name in r) and (s != _Stage.configure_called) and (s != _Stage.gp_created):
            raise AttributeError("Invalid write to field '%s'" % name)
        super().__setattr__(name, value)

    def __init__(self, generator_params: dict = {}):
        context = active_generator_context()

        self._target = context.target()
        self._autoscheduler = context.autoscheduler_params()
        self._stage = _Stage.generator_created
        self._gp_dict = None
        self._inputs_dict = None
        self._outputs_dict = None
        self._arginfos_in = None
        self._arginfos_out = None
        self._pipeline = None
        self._input_parameters = None
        self._output_funcs = None
        self._unhandled_generator_params = {}
        self._requirements = {}
        self._replacements = {}
        self._in_configure = 0
        self._pipeline_requirements = []

        self._advance_to_gp_created()
        if generator_params:
            _check(isinstance(generator_params, dict), "generator_params must be a dict")
            for k, v in generator_params.items():
                self._set_generatorparam_value(k, v)

    def allow_out_of_order_inputs_and_outputs(self):
        return False

    def configure(self):
        pass

    @abstractmethod
    def generate(self):
        pass

    def _add_gpio(self, name: str, io, io_dict: dict, arginfos: Optional[list]) -> None:
        _check_valid_name(name)
        _check(
            name not in self._requirements,
            "The name '%s' is used more than once in Generator %s." % (name, self._get_name()),
        )
        types, dimensions = io._get_types_and_dimensions()
        types, dimensions = self._set_io_types_and_dimensions_from_gp(name, types, dimensions)
        r = Requirement(name, types, dimensions)
        self._requirements[name] = r
        io_dict[name] = io
        if arginfos is not None:
            arginfos.append(ArgInfo(name, *io._get_direction_and_kind(), types, dimensions))

    def add_input(self, name: str, io) -> None:
        _check(self._in_configure > 0, "Can only call add_input() from the configure() method.")
        _check(not hasattr(self, name),
               "Cannot call add_input('%s') because the class already has a member of that name." % name)
        _check(isinstance(io, (InputBuffer, InputScalar)),
               "Cannot call add_input() with an object of type '%s'." % type(io))
        self._add_gpio(name, io, self._inputs_dict, self._arginfos_in)

    def add_output(self, name: str, io) -> None:
        _check(self._in_configure > 0, "Can only call add_output() from the configure() method.")
        _check(not hasattr(self, name),
               "Cannot call add_output('%s') because the class already has a member of that name." % name)
        _check(isinstance(io, (OutputBuffer, OutputScalar)),
               "Cannot call add_output() with an object of type '%s'." % type(io))
        self._add_gpio(name, io, self._outputs_dict, self._arginfos_out)

    def _advance_to_stage(self, new_stage: _Stage):
        _stage_advancers = {
            _Stage.gp_created: self._advance_to_gp_created,
            _Stage.gp_replaced: self._advance_to_gp_replaced,
            _Stage.io_created: self._advance_to_io_created,
            _Stage.configure_called: self._advance_to_configure_called,
            _Stage.io_replaced: self._advance_to_io_replaced,
        }
        assert new_stage in _stage_advancers
        a = _stage_advancers[new_stage]
        old_stage = self._stage
        if self._stage < new_stage:
            a()
        assert self._stage >= new_stage

    def _advance_to_gp_created(self):
        assert self._stage < _Stage.gp_created
        assert not self._gp_dict
        assert not self._inputs_dict
        assert not self._outputs_dict

        self._gp_dict = {}
        for name, gp in _unsorted_cls_dir(self.__class__):
            if not isinstance(gp, GeneratorParam):
                continue
            self._add_gpio(name, gp, self._gp_dict, None)

        self._stage = _Stage.gp_created

        if hasattr(self, "_halide_alias_generator_params"):
            for k, v in self._halide_alias_generator_params.items():
                self._set_generatorparam_value(k, v)

    def _advance_to_io_created(self):
        self._advance_to_stage(_Stage.gp_replaced)
        assert self._gp_dict is not None
        assert not self._inputs_dict
        assert not self._outputs_dict

        self._inputs_dict = {}
        self._outputs_dict = {}
        self._arginfos_in = []
        self._arginfos_out = []
        outputs_seen = False
        for name, io in _unsorted_cls_dir(self.__class__):
            is_input = isinstance(io, (InputBuffer, InputScalar))
            is_output = isinstance(io, (OutputBuffer, OutputScalar))
            if not (is_input or is_output):
                continue

            if is_input and outputs_seen and not self.allow_out_of_order_inputs_and_outputs():
                io_order_warning = ("Generators will always produce code that orders all Inputs before all Outputs; "
                                   "this Generator declares the Inputs and Outputs in a different order, so the calling convention may not be as expected. "
                                   "A future version of Halide will make this illegal, and require all Inputs to be declared before all Outputs. "
                                   "(You can avoid this requirement by overriding Generator::allow_out_of_order_inputs_and_outputs().)")
                warnings.warn(io_order_warning)

            if is_output:
                outputs_seen = True
            d = self._inputs_dict if is_input else self._outputs_dict
            a = self._arginfos_in if is_input else self._arginfos_out
            self._add_gpio(name, io, d, a)

        _check(
            len(self._unhandled_generator_params) == 0,
            "Generator %s has no GeneratorParam(s) named: %s" %
            (self._get_name(), list(self._unhandled_generator_params.keys())),
        )

        self._stage = _Stage.io_created

    def _set_io_types_and_dimensions_from_gp(self, name: str, current_types: list[Type],
                                             current_dimensions: int) -> (list[Type], int):
        new_types = current_types
        new_dimensions = current_dimensions

        type_name = "%s.type" % name
        if type_name in self._unhandled_generator_params:
            value = self._unhandled_generator_params.pop(type_name)
            new_types = GeneratorParam._parse_value(type_name, Type, value)
            _check(
                len(current_types) == 0,
                "Cannot set the GeneratorParam %s for %s because the value is explicitly specified in the Python source."
                % (type_name, self._get_name()),
            )

        dimensions_name = "%s.dim" % name
        if dimensions_name in self._unhandled_generator_params:
            value = self._unhandled_generator_params.pop(dimensions_name)
            new_dimensions = GeneratorParam._parse_value(dimensions_name, int, value)
            _check(
                current_dimensions == -1,
                "Cannot set the GeneratorParam %s for %s because the value is explicitly specified in the Python source."
                % (dimensions_name, self._get_name()),
            )

        return new_types, new_dimensions

    def _advance_to_configure_called(self):
        self._advance_to_stage(_Stage.io_created)

        self._in_configure += 1
        self.configure()
        self._in_configure -= 1

        _check(
            len(self._outputs_dict) > 0,
            "Generator '%s' must declare at least one output." % self._get_name(),
        )

        self._stage = _Stage.configure_called

    def _replace_one(self, gpio_dict):
        for name, gpio in gpio_dict.items():
            # Note that new_io will be a different type (e.g. InputBuffer -> ImageParam)
            new_gpio = self._replacements.pop(name, None)
            if not new_gpio:
                new_gpio = gpio._make_replacement(None, self._requirements[name])
            setattr(self, name, new_gpio)

    def _advance_to_gp_replaced(self):
        self._advance_to_stage(_Stage.gp_created)
        self._replace_one(self._gp_dict)
        self._stage = _Stage.gp_replaced

    def _advance_to_io_replaced(self):
        self._advance_to_stage(_Stage.configure_called)
        self._replace_one(self._inputs_dict)
        self._replace_one(self._outputs_dict)
        self._stage = _Stage.io_replaced

    # --------------- AbstractGenerator methods (partial)
    def _get_name(self) -> str:
        return self._halide_registered_name

    def _get_arginfos(self) -> list[ArgInfo]:
        self._advance_to_stage(_Stage.configure_called)
        return self._arginfos_in + self._arginfos_out

    def _set_generatorparam_value(self, name: str, value: Any):
        _check(
            name != "target",
            "The GeneratorParam named %s cannot be set by set_generatorparam_value." % name,
        )
        assert self._stage == _Stage.gp_created
        assert not self._pipeline
        if name in self._gp_dict:
            gp = self._gp_dict[name]
            assert isinstance(gp, GeneratorParam)
            old_value = gp._value
            new_value = GeneratorParam._parse_value(name, type(old_value), value)
            # Do not mutate the existing GP in place; it could be shared across multiple Generators.
            self._gp_dict[name] = GeneratorParam(new_value)
        elif name == "autoscheduler":
            _check(not self.autoscheduler().name, "The GeneratorParam %s cannot be set more than once" % name)
            self.autoscheduler().name = value
        elif name.startswith("autoscheduler."):
            sub_key = name[len("autoscheduler."):]
            _check(sub_key not in self.autoscheduler().extra,
                   "The GeneratorParam %s cannot be set more than once" % name)
            self.autoscheduler().extra[sub_key] = value
        else:
            self._unhandled_generator_params[name] = value

    def _build_pipeline(self) -> Pipeline:
        self._advance_to_stage(_Stage.io_replaced)

        assert not self._pipeline
        assert not self._input_parameters
        assert not self._output_funcs

        # Ensure that the current context is the one in self.
        # For most Generators this won't matter, but if the Generator
        # invokes SomeOtherGenerator.call(), it would be nice to have this
        # be the default, so that the end user doesn't have to mess with it.
        with self.context():
            self.generate()

        self._input_parameters = {n: getattr(self, n).parameter() for n in self._inputs_dict}
        self._output_funcs = {n: getattr(self, n) for n in self._outputs_dict}
        _check(
            len(self._output_funcs) > 0,
            "Generator %s must declare at least one Output in Arguments." % self._get_name(),
        )

        funcs = []
        for name, f in self._output_funcs.items():
            _check(f.defined(), "Output '%s' was not defined." % name)
            self._requirements[name]._check_types_and_dimensions(f.types(), f.dimensions())
            funcs.append(f)

        self._pipeline = Pipeline(funcs)
        for condition, error_args in self._pipeline_requirements:
            self._pipeline.add_requirement(condition, *error_args)
        self._stage = _Stage.pipeline_built
        return self._pipeline

    def _get_input_parameter(self, name: str) -> InternalParameter:
        assert self._stage == _Stage.pipeline_built
        _check(name in self._input_parameters, "Unknown input: %s" % name)
        return self._input_parameters[name]

    def _get_output_func(self, name: str) -> Func:
        assert self._stage == _Stage.pipeline_built
        _check(name in self._output_funcs, "Unknown output: %s" % name)
        return self._output_funcs[name]

    def _bind_input(self, name: str, values: list[object]):
        assert self._stage < _Stage.io_replaced
        self._advance_to_stage(_Stage.configure_called)
        _check(len(values) == 1, "Too many values specified for input: %s" % name)
        _check(name in self._inputs_dict, "There is no input with the name: %s" % name)
        assert name not in self._replacements
        self._replacements[name] = self._inputs_dict[name]._make_replacement(values[0], self._requirements[name])


_python_generators: dict = {}


def _get_python_generator_names() -> list[str]:
    return _python_generators.keys()


def _create_python_generator(name: str, context: GeneratorContext):
    cls = _python_generators.get(name, None)
    if not isclass(cls):
        return None
    with context:
        return cls()


def _fqname(o):
    k = o
    m = k.__module__
    q = k.__qualname__
    if m == "__main__" or k == "builtins":
        return q
    return m + "." + q


def active_generator_context() -> GeneratorContext:
    context = _halide_generator_context.get()
    _check(isinstance(context, GeneratorContext), "There is no active GeneratorContext")
    return context


def _is_interactive_mode() -> bool:
    return hasattr(sys, 'ps1')


def _check_generator_name_in_use(n:str):
    if _is_interactive_mode():
        # In interactive mode, it's OK to redefine generators... in fact, it's really
        # annoying not to allow this (e.g. in Colab)
        #
        # (Want to debug? Include this:)
        # if n in _python_generators:
        #     builtins.print("REDEFINING ", n)
        return

    _check(n not in _python_generators, "The Generator name %s is already in use." % n)


def alias(**kwargs):

    def alias_impl(cls):
        for k, v in kwargs.items():
            _check_valid_name(k)
            _check(hasattr(cls, "_halide_registered_name"), "@alias can only be used in conjunction with @generator.")
            _check_generator_name_in_use(k)
            _check(type(v) is dict, "The Generator alias %s specifies something other than a dict." % k)
            new_cls = type(k, (cls,), {"_halide_registered_name": k, "_halide_alias_generator_params": v})
            _python_generators[k] = new_cls
        return cls

    return alias_impl

def generator(name:str=""):
    # This code relies on dicts preserving key-insertion order, which is only
    # guaranteed for all Python implementations as of v3.7.
    _check(sys.version_info >= (3, 7), "Halide Generators require Python 3.7 or later.")
    def generator_impl(cls):
        n = name if name else _fqname(cls)
        _check_generator_name_in_use(n)
        _check(isclass(cls), "@generator can only be used on classes.")
        # Allow (but don't require) explicit inheritance from hl.Generator;
        # static type checkers (e.g. pytype) can complain that the decorated class
        # uses inherited methods since it can't correctly infer the inheritance.
        if issubclass(cls, Generator):
            new_cls = type(cls.__name__, (cls,), {"_halide_registered_name": n})
        else:
            new_cls = type(cls.__name__, (cls, Generator), {"_halide_registered_name": n})
        _python_generators[n] = new_cls
        return new_cls

    return generator_impl

def funcs(names:str) -> tuple(Func):
    """Given a space-delimited string, create a Func for each substring and return as a tuple."""
    return (Func(n) for n in names.split(' '))


def vars(names:str) -> tuple(Var):
    """Given a space-delimited string, create a Var for each substring and return as a tuple."""
    return (Var(n) for n in names.split(' '))



#!/usr/bin/python3

# Halide tutorial lesson 4

# This lesson demonstrates how to follow what Halide is doing at runtime.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_04_debugging_2
# in a shell with the current directory at python_bindings/

import halide as hl


def main():

    gradient = hl.Func("gradient")
    x, y = hl.Var("x"), hl.Var("y")

    # We'll define our gradient function as before.
    gradient[x, y] = x + y

    # And tell Halide that we'd like to be notified of all
    # evaluations.
    gradient.trace_stores()

    # Realize the function over an 8x8 region.
    print("Evaluating gradient")
    output = gradient.realize([8, 8])

    # This will print out all the times gradient(x, y) gets
    # evaluated.

    # Now that we can snoop on what Halide is doing, let's try our
    # first scheduling primitive. We'll make a new version of
    # gradient that processes each scanline in parallel.
    parallel_gradient = hl.Func("parallel_gradient")
    parallel_gradient[x, y] = x + y

    # We'll also trace this function.
    parallel_gradient.trace_stores()

    # Things are the same so far. We've defined the algorithm, but
    # haven't said anything about how to schedule it. In general,
    # exploring different scheduling decisions doesn't change the code
    # that describes the algorithm.

    # Now we tell Halide to use a parallel for loop over the y
    # coordinate. On linux we run this using a thread pool and a task
    # queue. On os x we call into grand central dispatch, which does
    # the same thing for us.
    parallel_gradient.parallel(y)

    # This time the printfs should come out of order, because each
    # scanline is potentially being processed in a different
    # thread. The number of threads should adapt to your system, but
    # on linux you can control it manually using the environment
    # variable HL_NUMTHREADS.
    print("\nEvaluating parallel_gradient")
    parallel_gradient.realize([8, 8])

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



import halide as hl
import numpy as np


def test_tuple_select():
    x = hl.Var("x")
    y = hl.Var("y")

    # ternary select with Expr condition
    f = hl.Func("f")
    f[x, y] = hl.select(x + y < 30, (x, y), (x - 1, y - 2))

    a, b = f.realize([200, 200])
    for xx in range(a.height()):
        for yy in range(a.width()):
            correct_a = xx if xx + yy < 30 else xx - 1
            correct_b = yy if xx + yy < 30 else yy - 2
            assert a[xx, yy] == correct_a
            assert b[xx, yy] == correct_b

    # ternary select with Tuple condition
    f = hl.Func("f")
    f[x, y] = hl.select((x < 30, y < 30), (x, y), (x - 1, y - 2))

    a, b = f.realize([200, 200])
    for xx in range(a.height()):
        for yy in range(a.width()):
            correct_a = xx if xx < 30 else xx - 1
            correct_b = yy if yy < 30 else yy - 2
            assert a[xx, yy] == correct_a
            assert b[xx, yy] == correct_b

    # multiway select with Expr condition
    f = hl.Func("f")
    # fmt: off
    f[x, y] = hl.select(x + y < 30,  (x, y),
                        x + y < 100, (x-1, y-2),
                                     (x-100, y-200))
    # fmt: on

    a, b = f.realize([200, 200])
    for xx in range(a.height()):
        for yy in range(a.width()):
            correct_a = xx if xx + yy < 30 else xx - 1 if xx + yy < 100 else xx - 100
            correct_b = yy if xx + yy < 30 else yy - 2 if xx + yy < 100 else yy - 200
            assert a[xx, yy] == correct_a
            assert b[xx, yy] == correct_b

    # multiway select with Tuple condition
    f = hl.Func("f")
    # fmt: off
    f[x, y] = hl.select((x < 30, y < 30),   (x, y),
                        (x < 100, y < 100), (x-1, y-2),
                                            (x-100, y-200))
    # fmt: on

    a, b = f.realize([200, 200])
    for xx in range(a.height()):
        for yy in range(a.width()):
            correct_a = xx if xx < 30 else xx - 1 if xx < 100 else xx - 100
            correct_b = yy if yy < 30 else yy - 2 if yy < 100 else yy - 200
            assert a[xx, yy] == correct_a
            assert b[xx, yy] == correct_b

    # Failure case: mixing Expr and Tuple in multiway
    try:
        f = hl.Func("f")
        # fmt: off
        f[x, y] = hl.select((x < 30, y < 30), (x, y),
                             x + y < 100,     (x-1, y-2),
                                              (x-100, y-200))
        # fmt: on
    except hl.HalideError as e:
        assert (
            "select() may not mix Expr and Tuple for the condition elements."
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"

    # Failure case: Tuples of mixed sizes
    try:
        f = hl.Func("f")
        f[x, y] = hl.select((x < 30, y < 30), (x, y, 0), (1, 2, 3, 4))
    except hl.HalideError as e:
        assert "select() on Tuples requires all Tuples to have identical sizes." in str(e)
    else:
        assert False, "Did not see expected exception!"


if __name__ == "__main__":
    test_tuple_select()



#!/usr/bin/python3

# Halide tutorial lesson 3

# This lesson demonstrates how to inspect what the Halide compiler is
# producing.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_03_debugging_1
# in a shell with the current directory at python_bindings/

import halide as hl


def main():

    # We'll start by defining the simple single-stage imaging
    # pipeline from lesson 1.

    # This lesson will be about debugging, but unfortunately in C++,
    # objects don't know their own names, which makes it hard for us
    # to understand the generated code. To get around this, you can
    # pass a string to the hl.Func and hl.Var constructors to give them a
    # name for debugging purposes.
    gradient = hl.Func("gradient")
    x, y = hl.Var("x"), hl.Var("y")
    gradient[x, y] = x + y

    # Realize the function to produce an output image. We'll keep it
    # very small for this lesson.
    output = gradient.realize([8, 8])

    # That line compiled and ran the pipeline. Try running this
    # lesson with the environment variable HL_DEBUG_CODEGEN set to
    # 1. It will print out the various stages of compilation, and a
    # pseudocode representation of the final pipeline.

    # If you set HL_DEBUG_CODEGEN to a higher number, you can see
    # more and more details of how Halide compiles your pipeline.
    # Setting HL_DEBUG_CODEGEN=2 shows the Halide code at each stage
    # of compilation, and also the llvm bitcode we generate at the
    # end.

    # Halide will also output an HTML version of this output, which
    # supports syntax highlighting and code-folding, so it can be
    # nicer to read for large pipelines. Open gradient.stmt.html" with your
    # browser after running this tutorial.
    gradient.compile_to_lowered_stmt("gradient.stmt.html", [], hl.StmtOutputFormat.HTML)

    # You can usually figure out what code Halide is generating using
    # this pseudocode. In the next lesson we'll see how to snoop on
    # Halide at runtime.

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Halide tutorial lesson 10.

# This lesson demonstrates how to use Halide as an more traditional
# ahead-of-time (AOT) compiler.

# This lesson is split across two files. The first (this one), builds
# a Halide pipeline and compiles it to an object file, a header and
# a Python extension. The second (lesson_10_aot_compilation_run.py),
# uses that object file to actually run the pipeline. This means that
# compiling this code is a multi-step process.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_10_aot_compilation_generate
# in a shell with the current directory at python_bindings/

# This will generate a file lesson_10_halide.py.cpp that still needs
# to be compiled. Use
#    make make test_tutorial_lesson_10_aot_compilation_run
# to generate and run a Python module called "lesson_10_halide".

# The benefits of this approach are that the final program:
# - Doesn't do any jit compilation at runtime, so it's fast.
# - Doesn't depend on libHalide at all, so it's a small, easy-to-deploy binary.

import halide as hl


def main():

    # We'll define a simple one-stage pipeline:
    brighter = hl.Func("brighter")
    x, y = hl.Var("x"), hl.Var("y")

    # The pipeline will depend on one scalar parameter.
    offset = hl.Param(hl.UInt(8), name="offset")

    # And take one grayscale 8-bit input buffer. The first
    # constructor argument gives the type of a pixel, and the second
    # specifies the number of dimensions (not the number of
    # channels!). For a grayscale image this is two for a color
    # image it's three. Currently, four dimensions is the maximum for
    # inputs and outputs.
    input = hl.ImageParam(hl.UInt(8), 2)

    # If we were jit-compiling, these would just be an int and a
    # hl.Buffer, but because we want to compile the pipeline once and
    # have it work for any value of the parameter, we need to make a
    # hl.Param object, which can be used like an hl.Expr, and an hl.ImageParam
    # object, which can be used like a hl.Buffer.

    # Define the hl.Func.
    brighter[x, y] = input[x, y] + offset

    # Schedule it.
    brighter.vectorize(x, 16).parallel(y)

    # This time, instead of calling brighter.realize(...), which
    # would compile and run the pipeline immediately, we'll call a
    # method that compiles the pipeline to an object file and header.
    #
    # For AOT-compiled code, we need to explicitly declare the
    # arguments to the routine. This routine takes two. Arguments are
    # usually Params or ImageParams.
    fname = "lesson_10_halide"
    brighter.compile_to({hl.OutputFileType.object: "lesson_10_halide.o",
                         hl.OutputFileType.python_extension: "lesson_10_halide.py.cpp"},
                        [input, offset], "lesson_10_halide")

    print("Halide pipeline compiled, but not yet run.")

    # To continue this lesson, look in the file
    # lesson_10_aot_compilation_run.py

    return 0


if __name__ == "__main__":
    main()



import halide as hl


def test_type():
    t1 = hl.Type()
    assert t1.code() == hl.TypeCode.Handle
    assert t1.bits() == 0
    assert t1.lanes() == 0

    t1 = hl.Type(hl.TypeCode.Int, 32, 1)
    t2 = hl.Int(32)
    assert t1 == t2
    assert t2.code() == hl.TypeCode.Int
    assert t2.bits() == 32
    assert t2.lanes() == 1
    assert t2.bytes() == 4

    t1 = t2.with_code(hl.TypeCode.UInt)
    assert t1 == hl.UInt(32)

    t1 = t2.with_bits(16)
    assert t1 == hl.Int(16)
    assert t1 != t2

    t1 = t2.with_lanes(8)
    assert t1 == hl.Int(32, 8)
    assert t1 != t2
    assert t1.element_of() == hl.Int(32)

    b1 = hl.Bool()
    f32 = hl.Float(32)
    h64 = hl.Handle()
    i32 = hl.Int(32)
    u32 = hl.UInt(32)
    vi32x8 = hl.Int(32, 8)
    u64 = hl.UInt(64)
    i64 = hl.Int(64)

    assert b1.is_bool()
    assert f32.is_float()
    assert h64.is_handle()
    assert i32.is_int()
    assert u32.is_uint()
    assert i64.is_int()
    assert u64.is_uint()

    assert not vi32x8.is_scalar()
    assert vi32x8.is_vector()

    h2 = hl.Handle()
    assert h64.same_handle_type(h2)

    assert f32.can_represent(b1)
    assert not b1.can_represent(f32)

    assert not b1.is_max(2)
    assert b1.is_max(1)
    assert b1.is_min(0)
    assert not b1.is_min(-1)

    assert i32.is_max(2147483647)
    assert i32.is_min(-2147483648)

    assert not u32.is_max(4294967296)
    assert u32.is_max(4294967295)
    assert u32.is_min(0)
    assert not u32.is_min(-1)
    assert not u32.is_min(1)

    assert i64.is_max(9223372036854775807)
    assert i64.is_min(-9223372036854775808)

    # Python doesn't have unsigned integers, so we can't really express this value.
    # assert u64.is_max(0xFFFFFFFFFFFFFFFF)
    assert u64.is_min(0)

    # repr() and str()
    assert str(i32) == "int32"
    assert repr(i32) == "<halide.Type int32>"
    assert str(vi32x8) == "int32x8"
    assert repr(vi32x8) == "<halide.Type int32x8>"


if __name__ == "__main__":
    test_type()



#!/usr/bin/python3

# Halide tutorial lesson 7

# This lesson demonstrates how express multi-stage pipelines.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_07_multi_stage_pipelines
# in a shell with the current directory at python_bindings/

import halide as hl

import halide.imageio
import numpy as np
import os.path


def main():
    # First we'll declare some Vars to use below.
    x, y, c = hl.Var("x"), hl.Var("y"), hl.Var("c")

    image_path = os.path.join(os.path.dirname(__file__), "../../tutorial/images/rgb.png")

    # Now we'll express a multi-stage pipeline that blurs an image
    # first horizontally, and then vertically.
    if True:
        # Take a color 8-bit input
        input = hl.Buffer(halide.imageio.imread(image_path))
        assert input.type() == hl.UInt(8)

        # Upgrade it to 16-bit, so we can do math without it overflowing.
        input_16 = hl.Func("input_16")
        input_16[x, y, c] = hl.cast(hl.UInt(16), input[x, y, c])

        # Blur it horizontally:
        blur_x = hl.Func("blur_x")
        blur_x[x, y, c] = (input_16[x - 1, y, c] + 2 *
                           input_16[x, y, c] + input_16[x + 1, y, c]) / 4

        # Blur it vertically:
        blur_y = hl.Func("blur_y")
        blur_y[x, y, c] = (blur_x[x, y - 1, c] + 2 *
                           blur_x[x, y, c] + blur_x[x, y + 1, c]) / 4

        # Convert back to 8-bit.
        output = hl.Func("output")
        output[x, y, c] = hl.cast(hl.UInt(8), blur_y[x, y, c])

        # Each hl.Func in this pipeline calls a previous one using
        # familiar function call syntax (we've overloaded operator()
        # on hl.Func objects). A hl.Func may call any other hl.Func that has
        # been given a definition. This restriction prevents
        # pipelines with loops in them. Halide pipelines are always
        # feed-forward graphs of Funcs.

        # Now let's realize it...

        # result = output.realize([input.width(), input.height(), 3])

        # Except that the line above is not going to work. Uncomment
        # it to see what happens.

        # Realizing this pipeline over the same domain as the input
        # image requires reading pixels out of bounds in the input,
        # because the blur_x stage reaches outwards horizontally, and
        # the blur_y stage reaches outwards vertically. Halide
        # detects this by injecting a piece of code at the top of the
        # pipeline that computes the region over which the input will
        # be read. When it starts to run the pipeline it first runs
        # this code, determines that the input will be read out of
        # bounds, and refuses to continue. No actual bounds checks
        # occur in the inner loop that would be slow.
        #
        # So what do we do? There are a few options. If we realize
        # over a domain shifted inwards by one pixel, we won't be
        # asking the Halide routine to read out of bounds. We saw how
        # to do this in the previous lesson:
        result = hl.Buffer(hl.UInt(8), [input.width() - 2, input.height() - 2, 3])
        result.set_min([1, 1])
        output.realize(result)

        # Save the result. It should look like a slightly blurry
        # parrot, and it should be two pixels narrower and two pixels
        # shorter than the input image.

        # python3-imageio versions <2.5 expect a numpy array
        halide.imageio.imwrite("blurry_parrot_1.png", np.asanyarray(result))
        print("Created blurry_parrot_1.png")

        # This is usually the fastest way to deal with boundaries:
        # don't write code that reads out of bounds :) The more
        # general solution is our next example.

    # The same pipeline, with a boundary condition on the input.
    if True:
        # Take a color 8-bit input
        input = hl.Buffer(halide.imageio.imread(image_path))
        assert input.type() == hl.UInt(8)

        # This time, we'll wrap the input in a hl.Func that prevents
        # reading out of bounds:
        clamped = hl.Func("clamped")

        # Define an expression that clamps x to lie within the the
        # range [0, input.width()-1].
        clamped_x = hl.clamp(x, 0, input.width() - 1)
        # Similarly hl.clamp y.
        clamped_y = hl.clamp(y, 0, input.height() - 1)
        # Load from input at the clamped coordinates. This means that
        # no matter how we evaluated the hl.Func 'clamped', we'll never
        # read out of bounds on the input. This is a hl.clamp-to-edge
        # style boundary condition, and is the simplest boundary
        # condition to express in Halide.
        clamped[x, y, c] = input[clamped_x, clamped_y, c]

        # Defining 'clamped' in that way can be done more concisely
        # using a helper function from the BoundaryConditions
        # namespace like so:
        #
        # clamped = hl.BoundaryConditions.repeat_edge(input)
        #
        # These are important to use for other boundary conditions,
        # because they are expressed in the way that Halide can best
        # understand and optimize.

        # Upgrade it to 16-bit, so we can do math without it
        # overflowing. This time we'll refer to our new hl.Func
        # 'clamped', instead of referring to the input image
        # directly.
        input_16 = hl.Func("input_16")
        input_16[x, y, c] = hl.cast(hl.UInt(16), clamped[x, y, c])

        # The rest of the pipeline will be the same...

        # Blur it horizontally:
        blur_x = hl.Func("blur_x")
        blur_x[x, y, c] = (input_16[x - 1, y, c] + 2 *
                           input_16[x, y, c] + input_16[x + 1, y, c]) / 4

        # Blur it vertically:
        blur_y = hl.Func("blur_y")
        blur_y[x, y, c] = (blur_x[x, y - 1, c] + 2 *
                           blur_x[x, y, c] + blur_x[x, y + 1, c]) / 4

        # Convert back to 8-bit.
        output = hl.Func("output")
        output[x, y, c] = hl.cast(hl.UInt(8), blur_y[x, y, c])

        # This time it's safe to evaluate the output over the some
        # domain as the input, because we have a boundary condition.
        result = output.realize([input.width(), input.height(), 3])

        # Save the result. It should look like a slightly blurry
        # parrot, but this time it will be the same size as the
        # input.

        # python3-imageio versions <2.5 expect a numpy array
        halide.imageio.imwrite("blurry_parrot_2.png", np.asanyarray(result))
        print("Created blurry_parrot_2.png")

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



"""
Shell for running Local Laplacian.
"""

from local_laplacian import local_laplacian
from local_laplacian_Mullapudi2016 import local_laplacian_Mullapudi2016
import halide.imageio
import numpy as np
import sys
import timeit


def main():
    if len(sys.argv) < 6:
        print(
            "Usage: %s input.png input.png levels alpha beta output.png" % sys.argv[0]
        )
        print("e.g. %s input.png 8 1 1 output.png 10" % sys.argv[0])
        sys.exit(1)

    input_path = sys.argv[1]
    levels = int(sys.argv[2])
    alpha = float(sys.argv[3])
    beta = float(sys.argv[4])
    output_path = sys.argv[5]
    timing_iterations = 10

    print("Reading from %s ..." % input_path)
    input_buf_u8 = halide.imageio.imread(input_path)
    assert input_buf_u8.dtype == np.uint8
    # Convert to uint16 in range [0..1]
    input_buf = input_buf_u8.astype(np.uint16) * 257
    h = input_buf.shape[1]
    w = input_buf.shape[2]
    output_buf = np.empty([3, h, w], dtype=input_buf.dtype)

    tests = {
        "Manual": local_laplacian,
        "Mullapudi2016": local_laplacian_Mullapudi2016,
    }

    for name, fn in tests.items():
        print("Running %s... " % name, end="")
        t = timeit.Timer(lambda: fn(input_buf, levels, alpha / (levels - 1), beta, output_buf))
        avg_time_sec = t.timeit(number=timing_iterations) / timing_iterations
        print("time: %fms" % (avg_time_sec * 1e3))

    output_buf_u8 = (output_buf // 257).astype(np.uint8)

    print("Saving to %s ..." % output_path)
    halide.imageio.imwrite(output_path, output_buf_u8)

    print("Success!")
    sys.exit(0)


if __name__ == "__main__":
    main()



import halide as hl

test_exterior = 42
test_min = -25
test_extent = 100

x, y = hl.vars("x y")


def expect_eq(actual, expected):
    assert expected == actual, "Failed: expected %d, actual %d" % (expected, actual)


def schedule_test(f, vector_width, target, partition_policy):
    if vector_width != 1:
        f.vectorize(x, vector_width)

    f.partition(x, partition_policy);
    f.partition(y, partition_policy);

    if target.has_gpu_feature() and vector_width <= 16:
        xo, yo, xi, yi = hl.vars("xo yo xi yi")
        f.gpu_tile(x, y, xo, yo, xi, yi, 2, 2)


def realize_and_check(
    f,
    checker,
    input,
    test_min_x,
    test_extent_x,
    test_min_y,
    test_extent_y,
    vector_width,
    target,
    partition_policy,
):
    result = hl.Buffer(hl.UInt(8), [test_extent_x, test_extent_y])
    result.set_min([test_min_x, test_min_y])
    f2 = hl.lambda_func(x, y, f[x, y])
    schedule_test(f2, vector_width, target, partition_policy)
    f2.realize(result, target)
    result.copy_to_host()
    for r in range(test_min_y, test_min_y + test_extent_y):
        for c in range(test_min_x, test_min_x + test_extent_x):
            checker(input, result, c, r)


def check_constant_exterior(input, result, c, r):
    if c < 0 or r < 0 or c >= input.width() or r >= input.height():
        expect_eq(result[c, r], test_exterior)
    else:
        expect_eq(result[c, r], input[c, r])


def check_repeat_edge(input, result, c, r):
    clamped_y = min(input.height() - 1, max(0, r))
    clamped_x = min(input.width() - 1, max(0, c))
    expect_eq(result[c, r], input[clamped_x, clamped_y])


def check_repeat_image(input, result, c, r):
    mapped_x = c
    mapped_y = r
    while mapped_x < 0:
        mapped_x += input.width()
    while mapped_x > input.width() - 1:
        mapped_x -= input.width()
    while mapped_y < 0:
        mapped_y += input.height()
    while mapped_y > input.height() - 1:
        mapped_y -= input.height()
    expect_eq(result[c, r], input[mapped_x, mapped_y])


def check_mirror_image(input, result, c, r):
    mapped_x = -(c + 1) if c < 0 else c
    mapped_x = mapped_x % (2 * input.width())
    if mapped_x > (input.width() - 1):
        mapped_x = (2 * input.width() - 1) - mapped_x
    mapped_y = -(r + 1) if r < 0 else r
    mapped_y = mapped_y % (2 * input.height())
    if mapped_y > (input.height() - 1):
        mapped_y = (2 * input.height() - 1) - mapped_y
    expect_eq(result[c, r], input[mapped_x, mapped_y])


def check_mirror_interior(input, result, c, r):
    mapped_x = abs(c) % (input.width() * 2 - 2)
    if mapped_x > input.width() - 1:
        mapped_x = input.width() * 2 - 2 - mapped_x
    mapped_y = abs(r) % (input.height() * 2 - 2)
    if mapped_y > input.height() - 1:
        mapped_y = input.height() * 2 - 2 - mapped_y
    expect_eq(result[c, r], input[mapped_x, mapped_y])


def test_all(vector_width, target, partition_policy):
    # print("target is %s, partition_policy is %s " % (str(target), str(partition_policy)))

    W = 32
    H = 32
    input = hl.Buffer(hl.UInt(8), [W, H])
    for r in range(H):
        for c in range(W):
            input[c, r] = (c + r * W) & 0xFF

    input_f = hl.Func()
    input_f[x, y] = input[x, y]

    tests = [
        (hl.BoundaryConditions.constant_exterior, check_constant_exterior),
        (hl.BoundaryConditions.repeat_edge, check_repeat_edge),
        (hl.BoundaryConditions.repeat_image, check_repeat_image),
        (hl.BoundaryConditions.mirror_image, check_mirror_image),
        (hl.BoundaryConditions.mirror_interior, check_mirror_interior),
    ]

    for bc, checker in tests:
        # print('  Testing %s:%d...' % (bc.__name__, vector_width))
        func_input_args = {"f": input_f, "bounds": [(0, W), (0, H)]}
        image_input_args = {"f": input, "bounds": [(0, W), (0, H)]}
        undef_min_args = {"f": input, "bounds": [(hl.Expr(), hl.Expr()), (0, H)]}
        undef_max_args = {"f": input, "bounds": [(0, W), (hl.Expr(), hl.Expr())]}
        implicit_bounds_args = {"f": input}

        if bc == hl.BoundaryConditions.constant_exterior:
            func_input_args["exterior"] = test_exterior
            image_input_args["exterior"] = test_exterior
            undef_min_args["exterior"] = test_exterior
            undef_max_args["exterior"] = test_exterior
            implicit_bounds_args["exterior"] = test_exterior

        realize_and_check(
            bc(**func_input_args),
            checker,
            input,
            test_min,
            test_extent,
            test_min,
            test_extent,
            vector_width,
            target,
            partition_policy,
        )
        realize_and_check(
            bc(**image_input_args),
            checker,
            input,
            test_min,
            test_extent,
            test_min,
            test_extent,
            vector_width,
            target,
            partition_policy,
        )
        realize_and_check(
            bc(**undef_min_args),
            checker,
            input,
            0,
            W,
            test_min,
            test_extent,
            vector_width,
            target,
            partition_policy,
        )
        realize_and_check(
            bc(**undef_max_args),
            checker,
            input,
            test_min,
            test_extent,
            0,
            H,
            vector_width,
            target,
            partition_policy,
        )
        realize_and_check(
            bc(**implicit_bounds_args),
            checker,
            input,
            test_min,
            test_extent,
            test_min,
            test_extent,
            vector_width,
            target,
            partition_policy,
        )


if __name__ == "__main__":
    target = hl.get_jit_target_from_environment()

    vector_width_power_max = 6
    # https://github.com/halide/Halide/issues/2148
    if target.has_feature(hl.TargetFeature.Metal) or \
        target.has_feature(hl.TargetFeature.Vulkan) or \
        target.has_feature(hl.TargetFeature.OpenGLCompute) or \
        target.has_feature(hl.TargetFeature.D3D12Compute):
        vector_width_power_max = 2

    for i in range(0, vector_width_power_max):
        vector_width = 1 << i
        test_all(vector_width, target, hl.Partition.Auto)
        test_all(vector_width, target, hl.Partition.Never)



import halide as hl
import sys

def main():
    x = hl.Var('x')
    f_in = hl.Func('in')
    f_in[x] = hl.f32(x) # Cast to float 32
    f_0 = hl.Func('f_0')
    f_0[x] = 2 * f_in[x]
    f_1 = hl.Func('f_1')
    f_1[x] = hl.sin(f_0[x])
    f_2 = hl.Func('f_2')
    f_2[x] = f_1[x] * f_1[x]

    # Setup
    f_2.set_estimate(x, 0, 1000)
    p = hl.Pipeline(f_2)
    target = hl.Target()
    asp = hl.AutoschedulerParams('Li2018', {'parallelism': 32})
    result = p.apply_autoscheduler(target, asp)
    print('Schedule:')
    print(result.schedule_source)

    p.compile_jit() # compile
    buf = p.realize([1000]) # compute and get the buffer

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: test path/to/Li2018-autoscheduler-plugin",sys.argv)
        sys.exit(1)

    hl.load_plugin(sys.argv[1])

    main()



#!/usr/bin/python3

# Halide tutorial lesson 12.

# This lesson demonstrates how to use Halide to run code on a GPU.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_12_using_the_gpu
# in a shell with the current directory at python_bindings/

import halide as hl

import halide.imageio
import os.path
import struct

# Include a clock to do performance testing.
from datetime import datetime

# Define some Vars to use.
x, y, c, i, xo, yo, xi, yi = hl.Var("x"), hl.Var("y"), hl.Var("c"), hl.Var(
    "i"), hl.Var("xo"), hl.Var("yo"), hl.Var("xi"), hl.Var("yi")

# We're going to want to schedule a pipeline in several ways, so we
# define the pipeline in a class so that we can recreate it several
# times with different schedules.


class MyPipeline:

    def __init__(self, input):

        assert input.type() == hl.UInt(8)

        self.lut = hl.Func("lut")
        self.padded = hl.Func("padded")
        self.padded16 = hl.Func("padded16")
        self.sharpen = hl.Func("sharpen")
        self.curved = hl.Func("curved")
        self.input = input

        # For this lesson, we'll use a two-stage pipeline that sharpens
        # and then applies a look-up-table (LUT).

        # First we'll define the LUT. It will be a gamma curve.
        gamma = hl.f32(1.2)
        self.lut[i] = hl.u8(hl.clamp(hl.pow(i / 255.0, gamma) * 255.0, 0, 255))

        # Augment the input with a boundary condition.
        self.padded[x, y, c] = input[hl.clamp(x, 0, input.width() - 1),
                                     hl.clamp(y, 0, input.height() - 1), c]

        # Cast it to 16-bit to do the math.
        self.padded16[x, y, c] = hl.u16(self.padded[x, y, c])

        # Next we sharpen it with a five-tap filter.
        self.sharpen[x, y, c] = (self.padded16[x, y, c] * 2 -
                                 (self.padded16[x - 1, y, c] +
                                  self.padded16[x, y - 1, c] +
                                  self.padded16[x + 1, y, c] +
                                  self.padded16[x, y + 1, c]) / 4)

        # Then apply the LUT.
        self.curved[x, y, c] = self.lut[self.sharpen[x, y, c]]

    # Now we define methods that give our pipeline several different
    # schedules.
    def schedule_for_cpu(self):
        # Compute the look-up-table ahead of time.
        self.lut.compute_root()

        # Compute color channels innermost. Promise that there will
        # be three of them and unroll across them.
        self.curved.reorder(c, x, y) \
            .bound(c, 0, 3) \
            .unroll(c)

        # Look-up-tables don't vectorize well, so just parallelize
        # curved in slices of 16 scanlines.
        self.curved.split(y, yo, yi, 16) \
            .parallel(yo)

        # Compute sharpen as needed per scanline of curved, reusing
        # previous values computed within the same strip of 16
        # scanlines.
        self.sharpen.compute_at(self.curved, yi)

        # Vectorize the sharpen. It's 16-bit so we'll vectorize it 8-wide.
        self.sharpen.vectorize(x, 8)

        # Compute the padded input at the same granularity as the
        # sharpen. We'll leave the cast to 16-bit inlined into
        # sharpen.
        self.padded.store_at(self.curved, yo) \
            .compute_at(self.curved, yi)

        # Also vectorize the padding. It's 8-bit, so we'll vectorize
        # 16-wide.
        self.padded.vectorize(x, 16)

        # JIT-compile the pipeline for the CPU.
        target = hl.get_host_target()
        self.curved.compile_jit(target)

        return

    # Now a schedule that uses CUDA or OpenCL.
    def schedule_for_gpu(self):
        target = find_gpu_target()
        if not target.has_gpu_feature():
            return False

        # If you want to see all of the OpenCL, Metal, CUDA or D3D 12 API
        # calls done by the pipeline, you can also enable the Debug flag.
        # This is helpful for figuring out which stages are slow, or when
        # CPU -> GPU copies happen. It hurts performance though, so we'll
        # leave it commented out.
        # target.set_feature(hl.TargetFeature.Debug)

        # We make the decision about whether to use the GPU for each
        # hl.Func independently. If you have one hl.Func computed on the
        # CPU, and the next computed on the GPU, Halide will do the
        # copy-to-gpu under the hood. For this pipeline, there's no
        # reason to use the CPU for any of the stages. Halide will
        # copy the input image to the GPU the first time we run the
        # pipeline, and leave it there to reuse on subsequent runs.

        # As before, we'll compute the LUT once at the start of the
        # pipeline.
        self.lut.compute_root()

        # Let's compute the look-up-table using the GPU in 16-wide
        # one-dimensional thread blocks. First we split the index
        # into blocks of size 16:
        block, thread = hl.Var("block"), hl.Var("thread")
        self.lut.split(i, block, thread, 16)
        # Then we tell cuda that our Vars 'block' and 'thread'
        # correspond to CUDA's notions of blocks and threads, or
        # OpenCL's notions of thread groups and threads.
        self.lut.gpu_blocks(block) \
                .gpu_threads(thread)

        # This is a very common scheduling pattern on the GPU, so
        # there's a shorthand for it:

        # lut.gpu_tile(i, block, thread, 16);

        # hl.Func.gpu_tile behaves the same as hl.Func.tile, except that
        # it also specifies that the tile coordinates correspond to
        # GPU blocks, and the coordinates within each tile correspond
        # to GPU threads.

        # Compute color channels innermost. Promise that there will
        # be three of them and unroll across them.
        self.curved.reorder(c, x, y) \
                   .bound(c, 0, 3) \
                   .unroll(c)

        # Compute curved in 2D 8x8 tiles using the GPU.
        self.curved.gpu_tile(x, y, xo, yo, xi, yi, 8, 8)

        # This is equivalent to:
        # curved.tile(x, y, xo, yo, xi, yi, 8, 8)
        #       .gpu_blocks(xo, yo)
        #       .gpu_threads(xi, yi)

        # We'll leave sharpen as inlined into curved.

        # Compute the padded input as needed per GPU block, storing
        # the intermediate result in shared memory. In the schedule
        # above xo corresponds to GPU blocks.
        self.padded.compute_at(self.curved, xo)

        # Use the GPU threads for the x and y coordinates of the
        # padded input.
        self.padded.gpu_threads(x, y)

        # JIT-compile the pipeline for the GPU. CUDA, OpenCL, or
        # Metal are not enabled by default. We have to construct a
        # Target object, enable one of them, and then pass that
        # target object to compile_jit. Otherwise your CPU will very
        # slowly pretend it's a GPU, and use one thread per output
        # pixel.

        print("Target: ", target)
        self.curved.compile_jit(target)
        return True

    def test_performance(self):
        # Test the performance of the scheduled MyPipeline.

        output = hl.Buffer(hl.UInt(8),
                           [self.input.width(), self.input.height(), self.input.channels()])

        # Run the filter once to initialize any GPU runtime state.
        self.curved.realize(output)

        # Now take the best of 3 runs for timing.
        best_time = float("inf")
        for i in range(3):

            t1 = datetime.now()

            # Run the filter 100 times.
            for j in range(100):
                self.curved.realize(output)

            # Force any GPU code to finish by copying the buffer back to the
            # CPU.
            output.copy_to_host()

            t2 = datetime.now()

            elapsed = (t2 - t1).total_seconds()
            if elapsed < best_time:
                best_time = elapsed
        # end of "best of three times"

        print("%1.4f milliseconds" % (best_time * 1000))

    def test_correctness(self, reference_output):

        assert reference_output.type() == hl.UInt(8)
        output = self.curved.realize([self.input.width(),
                                      self.input.height(),
                                      self.input.channels()])
        assert output.type() == hl.UInt(8)

        # Check against the reference output.
        for cc in range(self.input.channels()):
            for yy in range(self.input.height()):
                for xx in range(self.input.width()):
                    assert output[xx, yy, cc] == reference_output[xx, yy, cc], \
                        "Mismatch between output (%d) and reference output (%d) at %d, %d, %d" % (
                            output[xx, yy, cc], reference_output[xx, yy, cc], xx, yy, cc)

        print("CPU and GPU outputs are consistent.")


def main():
    # Load an input image.
    image_path = os.path.join(os.path.dirname(__file__), "../../tutorial/images/rgb.png")
    input = hl.Buffer(halide.imageio.imread(image_path))

    # Allocated an image that will store the correct output
    reference_output = hl.Buffer(hl.UInt(8), [input.width(), input.height(), input.channels()])

    print("Running pipeline on CPU:")
    p1 = MyPipeline(input)
    p1.schedule_for_cpu()
    p1.curved.realize(reference_output)

    print("Running pipeline on GPU:")
    p2 = MyPipeline(input)
    has_gpu_target = p2.schedule_for_gpu()
    if has_gpu_target:
        print("Testing GPU correctness:")
        p2.test_correctness(reference_output)
    else:
        print("No GPU target available on the host")

    print("Testing performance on CPU:")
    p1.test_performance()

    if has_gpu_target:
        print("Testing performance on GPU:")
        p2.test_performance()

    return 0

# A helper function to check if OpenCL, Metal or D3D12 is present on the
# host machine.


def find_gpu_target():
    # Start with a target suitable for the machine you're running this on.
    target = hl.get_host_target()

    features_to_try = []
    if target.os == hl.TargetOS.Windows:
        # Try D3D12 first; if that fails, try OpenCL.
        if struct.calcsize("P") == 8:
            # D3D12Compute support is only available on 64-bit systems at present.
            features_to_try.append(hl.TargetFeature.D3D12Compute)
        features_to_try.append(hl.TargetFeature.OpenCL)
    elif target.os == hl.TargetOS.OSX:
        # OS X doesn't update its OpenCL drivers, so they tend to be broken.
        # CUDA would also be a fine choice on machines with NVidia GPUs.
        features_to_try.append(hl.TargetFeature.Metal)
    else:
        features_to_try.append(hl.TargetFeature.OpenCL)

    # Uncomment the following lines to also try CUDA:
    # features_to_try.append(hl.TargetFeature.CUDA);
    for f in features_to_try:
        new_target = target.with_feature(f)
        if (hl.host_supports_target_device(new_target)):
            return new_target

    print("Requested GPU(s) are not supported. (Do you have the proper hardware and/or driver installed?)")
    return target


if __name__ == "__main__":
    main()



# Before reading this file, see lesson_10_aot_compilation_generate.py

# This is the code that actually uses the Halide pipeline we've
# compiled. It does not depend on libHalide, so we won't do
# "import halide".
#
# Instead, it depends on the header file that lesson_10_generate
# produced when we ran it:
import lesson_10_halide

import numpy as np


def main():
    # Have a look at the generated files above (they won't exist until you've run
    # lesson_10_generate): lesson_10_halide.py.cpp, lesson_10_halide.h
    #
    # In the header file, the generated function is represented like this:
    # int lesson_10_halide(halide_buffer_t*, uint8_t, halide_buffer_t*);
    #
    # lesson_10_halide.py.cpp creates a Python wrapper around this function.
    # Buffers are converted using the Python buffer API:
    #
    # https://docs.python.org/2/c-api/buffer.html
    # https://docs.python.org/3/c-api/buffer.html
    #
    # In other words, you can pass numpy arrays directly to the generated
    # code.

    # Let's make some input data to test with:
    input = np.empty((640, 480), dtype=np.uint8, order='F')
    for y in range(480):
        for x in range(640):
            input[x, y] = x ^ (y + 1)

    # And the memory where we want to write our output:
    output = np.empty((640, 480), dtype=np.uint8, order='F')

    offset_value = 5

    lesson_10_halide.lesson_10_halide(input, offset_value, output)

    # Now let's check the filter performed as advertised. It was
    # supposed to add the offset to every input pixel.
    correct_val = np.empty((1), dtype=np.uint8)
    for y in range(480):
        for x in range(640):
            input_val = input[x, y]
            output_val = output[x, y]
            correct_val[0] = input_val
            # we add over a uint8 value (will properly model overflow)
            correct_val[0] += offset_value
            assert output_val == correct_val[0], \
                "output(%d, %d) was %d instead of %d" % (x, y, output_val, correct_val)

    # Everything worked!
    print("Success!")
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Halide tutorial lesson 5

# This lesson demonstrates how to manipulate the order in which you
# evaluate pixels in a hl.Func, including vectorization,
# parallelization, unrolling, and tiling.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_05_scheduling_1
# in a shell with the current directory at python_bindings/
import halide as hl


def main():

    # We're going to define and schedule our gradient function in
    # several different ways, and see what order pixels are computed
    # in.

    x, y = hl.Var("x"), hl.Var("y")

    # First we observe the default ordering.
    if True:
        gradient = hl.Func("gradient")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # By default we walk along the rows and then down the
        # columns. This means x varies quickly, and y varies
        # slowly. x is the column and y is the row, so this is a
        # row-major traversal.
        print("Evaluating gradient row-major")
        output = gradient.realize([4, 4])

        # The equivalent C is:
        print("Equivalent C:")
        for yy in range(4):
            for xx in range(4):
                print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print("\n")

        # Tracing is one useful way to understand what a schedule is
        # doing. You can also ask Halide to print out pseudocode
        # showing what loops Halide is generating:
        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

        # Because we're using the default ordering, it should print:
        # compute gradient:
        #   for y:
        #     for x:
        #       gradient(...) = ...

    # Reorder variables.
    if True:
        gradient = hl.Func("gradient_col_major")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # If we reorder x and y, we can walk down the columns
        # instead. The reorder call takes the arguments of the func,
        # and sets a new nesting order for the for loops that are
        # generated. The arguments are specified from the innermost
        # loop out, so the following call puts y in the inner loop:
        gradient.reorder(y, x)

        # This means y (the row) will vary quickly, and x (the
        # column) will vary slowly, so this is a column-major
        # traversal.

        print("Evaluating gradient column-major")
        output = gradient.realize([4, 4])

        print("Equivalent C:")
        for yy in range(4):
            for xx in range(4):
                print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        # If we print pseudo-code for this schedule, we'll see that
        # the loop over y is now inside the loop over x.
        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Split a variable into two.
    if True:
        gradient = hl.Func("gradient_split")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # The most powerful primitive scheduling operation you can do
        # to a var is to split it into inner and outer sub-variables:
        x_outer, x_inner = hl.Var("x_outer"), hl.Var("x_inner")
        gradient.split(x, x_outer, x_inner, 2)

        # This breaks the loop over x into two nested loops: an outer
        # one over x_outer, and an inner one over x_inner. The last
        # argument to split was the "split factor". The inner loop
        # runs from zero to the split factor. The outer loop runs
        # from zero to the extent required of x (4 in this case)
        # divided by the split factor. Within the loops, the old
        # variable is defined to be outer * factor + inner. If the
        # old loop started at a value other than zero, then that is
        # also added within the loops.

        print("Evaluating gradient with x split into x_outer and x_inner ")
        output = gradient.realize([4, 4])

        print("Equivalent C:")
        for yy in range(4):
            for x_outer in range(2):
                for x_inner in range(2):
                    xx = x_outer * 2 + x_inner
                    print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

        # Note that the order of evaluation of pixels didn't actually
        # change! Splitting by itself does nothing, but it does open
        # up all of the scheduling possibilities that we will explore
        # below.

    # Fuse two variables into one.
    if True:
        gradient = hl.Func("gradient_fused")
        gradient[x, y] = x + y

        # The opposite of splitting is 'fusing'. Fusing two variables
        # merges the two loops into a single for loop over the
        # product of the extents. Fusing is less important than
        # splitting, but it also sees use (as we'll see later in this
        # lesson). Like splitting, fusing by itself doesn't change
        # the order of evaluation.
        fused = hl.Var("fused")
        gradient.fuse(x, y, fused)

        print("Evaluating gradient with x and y fused")
        output = gradient.realize([4, 4])

        print("Equivalent C:")
        for fused in range(4 * 4):
            yy = fused / 4
            xx = fused % 4
            print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Evaluating in tiles.
    if True:
        gradient = hl.Func("gradient_tiled")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # Now that we can both split and reorder, we can do tiled
        # evaluation. Let's split both x and y by a factor of two,
        # and then reorder the vars to express a tiled traversal.
        #
        # A tiled traversal splits the domain into small rectangular
        # tiles, and outermost iterates over the tiles, and within
        # that iterates over the points within each tile. It can be
        # good for performance if neighboring pixels use overlapping
        # input data, for example in a blur. We can express a tiled
        # traversal like so:
        x_outer, x_inner, y_outer, y_inner = hl.Var(), hl.Var(), hl.Var(), hl.Var()
        gradient.split(x, x_outer, x_inner, 2)
        gradient.split(y, y_outer, y_inner, 2)
        gradient.reorder(x_inner, y_inner, x_outer, y_outer)

        # This pattern is common enough that there's a shorthand for it:
        # gradient.tile(x, y, x_outer, y_outer, x_inner, y_inner, 2, 2)

        print("Evaluating gradient in 2x2 tiles")
        output = gradient.realize([4, 4])

        print("Equivalent C:")
        for y_outer in range(2):
            for x_outer in range(2):
                for y_inner in range(2):
                    for x_inner in range(2):
                        xx = x_outer * 2 + x_inner
                        yy = y_outer * 2 + y_inner
                        print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Evaluating in vectors.
    if True:
        gradient = hl.Func("gradient_in_vectors")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # The nice thing about splitting is that it guarantees the
        # inner variable runs from zero to the split factor. Most of
        # the time the split-factor will be a compile-time constant,
        # so we can replace the loop over the inner variable with a
        # single vectorized computation. This time we'll split by a
        # factor of four, because on X86 we can use SSE to compute in
        # 4-wide vectors.
        x_outer, x_inner = hl.Var("x_outer"), hl.Var("x_inner")
        gradient.split(x, x_outer, x_inner, 4)
        gradient.vectorize(x_inner)

        # Splitting and then vectorizing the inner variable is common
        # enough that there's a short-hand for it. We could have also
        # said:
        #
        # gradient.vectorize(x, 4)
        #
        # which is equivalent to:
        #
        # gradient.split(x, x, x_inner, 4)
        # gradient.vectorize(x_inner)
        #
        # Note that in this case we reused the name 'x' as the new
        # outer variable. Later scheduling calls that refer to x
        # will refer to this new outer variable named x.

        # This time we'll evaluate over an 8x4 box, so that we have
        # more than one vector of work per scanline.
        print("Evaluating gradient with x_inner vectorized ")
        output = gradient.realize([8, 4])

        print("Equivalent C:")
        for yy in range(4):
            for x_outer in range(2):
                # The loop over x_inner has gone away, and has been
                # replaced by a vectorized version of the
                # expression. On x86 processors, Halide generates SSE
                # for all of this.
                x_vec = [x_outer * 4 + 0,
                         x_outer * 4 + 1,
                         x_outer * 4 + 2,
                         x_outer * 4 + 3]
                val = [x_vec[0] + yy,
                       x_vec[1] + yy,
                       x_vec[2] + yy,
                       x_vec[3] + yy]
                print("Evaluating at <%d, %d, %d, %d>, <%d, %d, %d, %d>: <%d, %d, %d, %d>" % (
                    x_vec[0], x_vec[1], x_vec[2], x_vec[3], yy, yy, yy, yy, val[0], val[1], val[2], val[3]))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Unrolling a loop.
    if True:
        gradient = hl.Func("gradient_in_vectors")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # If multiple pixels share overlapping data, it can make
        # sense to unroll a computation so that shared values are
        # only computed or loaded once. We do this similarly to how
        # we expressed vectorizing. We split a dimension and then
        # fully unroll the loop of the inner variable. Unrolling
        # doesn't change the order in which things are evaluated.
        x_outer, x_inner = hl.Var("x_outer"), hl.Var("x_inner")
        gradient.split(x, x_outer, x_inner, 2)
        gradient.unroll(x_inner)

        # The shorthand for this is:
        # gradient.unroll(x, 2)

        print("Evaluating gradient unrolled by a factor of two")
        result = gradient.realize([4, 4])

        print("Equivalent C:")
        for yy in range(4):
            for x_outer in range(2):
                # Instead of a for loop over x_inner, we get two
                # copies of the innermost statement.
                if True:
                    x_inner = 0
                    xx = x_outer * 2 + x_inner
                    print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

                if True:
                    x_inner = 1
                    xx = x_outer * 2 + x_inner
                    print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Splitting by factors that don't divide the extent.
    if True:
        gradient = hl.Func("gradient_split_5x4")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # Splitting guarantees that the inner loop runs from zero to
        # the split factor, which is important for the uses we saw
        # above. So what happens when the total extent we wish to
        # evaluate x over isn't a multiple of the split factor? We'll
        # split by a factor of two again, but now we'll evaluate
        # gradient over a 5x4 box instead of the 4x4 box we've been
        # using.
        x_outer, x_inner = hl.Var("x_outer"), hl.Var("x_inner")
        gradient.split(x, x_outer, x_inner, 2)

        print("Evaluating gradient over a 5x4 box with x split by two ")
        output = gradient.realize([5, 4])

        print("Equivalent C:")
        for yy in range(4):
            for x_outer in range(3):  # Now runs from 0 to 3
                for x_inner in range(2):
                    xx = x_outer * 2
                    # Before we add x_inner, make sure we don't
                    # evaluate points outside of the 5x4 box. We'll
                    # hl.clamp x to be at most 3 (5 minus the split
                    # factor).
                    if xx > 3:
                        xx = 3
                    xx += x_inner
                    print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

        # If you read the output, you'll see that some coordinates
        # were evaluated more than once! That's generally OK, because
        # pure Halide functions have no side-effects, so it's safe to
        # evaluate the same point multiple times. If you're calling
        # out to C functions like we are, it's your responsibility to
        # make sure you can handle the same point being evaluated
        # multiple times.

        # The general rule is: If we require x from x_min to x_min + x_extent, and
        # we split by a factor 'factor', then:
        #
        # x_outer runs from 0 to (x_extent + factor - 1)/factor
        # x_inner runs from 0 to factor
        # x = hl.min(x_outer * factor, x_extent - factor) + x_inner + x_min
        #
        # In our example, x_min was 0, x_extent was 5, and factor was 2.

        # However, if you write a Halide function with an update
        # definition (see lesson 9), then it is not safe to evaluate
        # the same point multiple times, so we won't apply this
        # trick. Instead the range of values computed will be rounded
        # up to the next multiple of the split factor.

    # Fusing, tiling, and parallelizing.
    if True:
        # We saw in the previous lesson that we can parallelize
        # across a variable. Here we combine it with fusing and
        # tiling to express a useful pattern - processing tiles in
        # parallel.

        # This is where fusing shines. Fusing helps when you want to
        # parallelize across multiple dimensions without introducing
        # nested parallelism. Nested parallelism (parallel for loops
        # within parallel for loops) is supported by Halide, but
        # often gives poor performance compared to fusing the
        # parallel variables into a single parallel for loop.

        gradient = hl.Func("gradient_fused_tiles")
        gradient[x, y] = x + y
        gradient.trace_stores()

        # First we'll tile, then we'll fuse the tile indices and
        # parallelize across the combination.
        x_outer, y_outer = hl.Var("x_outer"), hl.Var("y_outer")
        x_inner, y_inner = hl.Var("x_inner"), hl.Var("y_inner")
        tile_index = hl.Var("tile_index")
        gradient.tile(x, y, x_outer, y_outer, x_inner, y_inner, 2, 2)
        gradient.fuse(x_outer, y_outer, tile_index)
        gradient.parallel(tile_index)

        # The scheduling calls all return a reference to the hl.Func, so
        # you can also chain them together into a single statement to
        # make things slightly clearer:
        #
        # gradient
        #     .tile(x, y, x_outer, y_outer, x_inner, y_inner, 2, 2)
        #     .fuse(x_outer, y_outer, tile_index)
        #     .parallel(tile_index)

        print("Evaluating gradient tiles in parallel")
        output = gradient.realize([4, 4])

        # The tiles should occur in arbitrary order, but within each
        # tile the pixels will be traversed in row-major order.

        print("Equivalent (serial) C:")
        # This outermost loop should be a parallel for loop, but that's hard in
        # C.
        for tile_index in range(4):
            y_outer = tile_index / 2
            x_outer = tile_index % 2
            for y_inner in range(2):
                for x_inner in range(2):
                    yy = y_outer * 2 + y_inner
                    xx = x_outer * 2 + x_inner
                    print("Evaluating at x = %d, y = %d: %d" % (xx, yy, xx + yy))

        print()

        print("Pseudo-code for the schedule:")
        gradient.print_loop_nest()
        print()

    # Putting it all together.
    if True:
        # Are you ready? We're going to use all of the features above now.
        gradient_fast = hl.Func("gradient_fast")
        gradient_fast[x, y] = x + y

        # We'll process 256x256 tiles in parallel.
        x_outer, y_outer = hl.Var("x_outer"), hl.Var("y_outer")
        x_inner, y_inner = hl.Var("x_inner"), hl.Var("y_inner")
        tile_index = hl.Var("tile_index")
        gradient_fast \
            .tile(x, y, x_outer, y_outer, x_inner, y_inner, 256, 256) \
            .fuse(x_outer, y_outer, tile_index) \
            .parallel(tile_index)

        # We'll compute two scanlines at once while we walk across
        # each tile. We'll also vectorize in x. The easiest way to
        # express this is to recursively tile again within each tile
        # into 4x2 subtiles, then vectorize the subtiles across x and
        # unroll them across y:
        x_inner_outer, y_inner_outer = hl.Var(
            "x_inner_outer"), hl.Var("y_inner_outer")
        x_vectors, y_pairs = hl.Var("x_vectors"), hl.Var("y_pairs")
        gradient_fast \
            .tile(x_inner, y_inner, x_inner_outer, y_inner_outer, x_vectors, y_pairs, 4, 2) \
            .vectorize(x_vectors) \
            .unroll(y_pairs)

        # Note that we didn't do any explicit splitting or
        # reordering. Those are the most important primitive
        # operations, but mostly they are buried underneath tiling,
        # vectorizing, or unrolling calls.

        # Now let's evaluate this over a range which is not a
        # multiple of the tile size.

        # If you like you can turn on tracing, but it's going to
        # produce a lot of prints. Instead we'll compute the answer
        # both in C and Halide and see if the answers match.
        result = gradient_fast.realize([800, 600])

        print("Checking Halide result against equivalent C...")
        for tile_index in range(4 * 3):
            y_outer = tile_index // 4
            x_outer = tile_index % 4
            for y_inner_outer in range(256 // 2):
                for x_inner_outer in range(256 // 4):
                    # We're vectorized across x
                    xx = min(x_outer * 256, 800 - 256) + x_inner_outer * 4
                    x_vec = [xx + 0,
                             xx + 1,
                             xx + 2,
                             xx + 3]

                    # And we unrolled across y
                    y_base = min(y_outer * 256, 600 - 256) + y_inner_outer * 2

                    if True:
                        # y_pairs = 0
                        yy = y_base + 0
                        y_vec = [yy, yy, yy, yy]
                        val = [x_vec[0] + y_vec[0],
                               x_vec[1] + y_vec[1],
                               x_vec[2] + y_vec[2],
                               x_vec[3] + y_vec[3]]

                        # Check the result.
                        for i in range(4):
                            assert result[x_vec[i], y_vec[i]] == val[i], \
                                "There was an error at %d %d!" % (x_vec[i], y_vec[i])

                    if True:
                        # y_pairs = 1
                        yy = y_base + 1
                        y_vec = [yy, yy, yy, yy]
                        val = [x_vec[0] + y_vec[0],
                               x_vec[1] + y_vec[1],
                               x_vec[2] + y_vec[2],
                               x_vec[3] + y_vec[3]]

                        # Check the result.
                        for i in range(4):
                            assert result[x_vec[i], y_vec[i]] == val[i], \
                                "There was an error at %d %d!" % (x_vec[i], y_vec[i])

        print()

        print("Pseudo-code for the schedule:")
        gradient_fast.print_loop_nest()
        print()

        # Note that in the Halide version, the algorithm is specified
        # once at the top, separately from the optimizations, and there
        # aren't that many lines of code total. Compare this to the C
        # version. There's more code (and it isn't even parallelized or
        # vectorized properly). More annoyingly, the statement of the
        # algorithm (the result is x plus y) is buried in multiple places
        # within the mess. This C code is hard to write, hard to read,
        # hard to debug, and hard to optimize further. This is why Halide
        # exists.

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



import halide as hl


def test_atomics():
    x = hl.Var("x")
    im = hl.Func("im")
    f = hl.Func("f")
    im[x] = (x * x) % 5
    r = hl.RDom([(0, 100)])
    f[x] = 0
    f[hl.Expr(im[r])] += 1
    f.compute_root().update().atomic().parallel(r)
    b = f.realize([5])

    ref = [0, 0, 0, 0, 0]
    for i in range(100):
        idx = (i * i) % 5
        ref[idx] += 1
    for i in range(5):
        assert b[i] == ref[i]


if __name__ == "__main__":
    test_atomics()



from addconstantcpp import addconstantcpp
from addconstantpy import addconstantpy
from addconstantcpp_with_offset_42 import addconstantcpp_with_offset_42
from addconstantpy_with_offset_42 import addconstantpy_with_offset_42
from addconstantcpp_with_negative_offset import addconstantcpp_with_negative_offset
from addconstantpy_with_negative_offset import addconstantpy_with_negative_offset
import numpy

TESTS_AND_OFFSETS = [
    (addconstantcpp, 0),
    (addconstantpy, 0),
    (addconstantcpp_with_offset_42, 42),
    (addconstantpy_with_offset_42, 42),
    (addconstantcpp_with_negative_offset, -1),
    (addconstantpy_with_negative_offset, -1),
]

ERROR_THRESHOLD = 0.0001


def test(addconstant_impl_func, offset):
    scalar_u1 = True
    scalar_u8 = 3
    scalar_u16 = 49153
    scalar_u32 = 65537
    scalar_u64 = 5724968371
    scalar_i8 = -7
    scalar_i16 = -30712
    scalar_i32 = 98901
    scalar_i64 = -8163465847
    scalar_float = 3.14159
    scalar_double = 1.61803

    input_u8 = numpy.array([0, 1, 2], dtype=numpy.uint8)
    input_u16 = numpy.array([0, 256, 512], dtype=numpy.uint16)
    input_u32 = numpy.array([0, 65536, 131072], dtype=numpy.uint32)
    input_u64 = numpy.array([0, 4294967296, 8589934592], dtype=numpy.uint64)
    input_i8 = numpy.array([1, -2, 3], dtype=numpy.int8)
    input_i16 = numpy.array([1, -256, 512], dtype=numpy.int16)
    input_i32 = numpy.array([1, -65536, 131072], dtype=numpy.int32)
    input_i64 = numpy.array([0, -4294967296, 8589934592], dtype=numpy.int64)
    input_float = numpy.array([3.14, 2.718, 1.618], dtype=numpy.float32)
    input_double = numpy.array([3.14, 2.718, 1.618], dtype=numpy.float64)
    input_half = numpy.array([3.14, 2.718, 1.618], dtype=numpy.float16)
    input_2d = numpy.array([[1, 2, 3], [4, 5, 6]], dtype=numpy.int8, order="F")
    input_3d = numpy.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=numpy.int8)

    output_u8 = numpy.zeros((3,), dtype=numpy.uint8)
    output_u16 = numpy.zeros((3,), dtype=numpy.uint16)
    output_u32 = numpy.zeros((3,), dtype=numpy.uint32)
    output_u64 = numpy.zeros((3,), dtype=numpy.uint64)
    output_i8 = numpy.zeros((3,), dtype=numpy.int8)
    output_i16 = numpy.zeros((3,), dtype=numpy.int16)
    output_i32 = numpy.zeros((3,), dtype=numpy.int32)
    output_i64 = numpy.zeros((3,), dtype=numpy.int64)
    output_float = numpy.zeros((3,), dtype=numpy.float32)
    output_double = numpy.zeros((3,), dtype=numpy.float64)
    output_half = numpy.zeros((3,), dtype=numpy.float16)
    output_2d = numpy.zeros((2, 3), dtype=numpy.int8, order="F")
    output_3d = numpy.zeros((2, 2, 2), dtype=numpy.int8)

    addconstant_impl_func(
        scalar_u1,
        scalar_u8,
        scalar_u16,
        scalar_u32,
        scalar_u64,
        scalar_i8,
        scalar_i16,
        scalar_i32,
        scalar_i64,
        scalar_float,
        scalar_double,
        input_u8,
        input_u16,
        input_u32,
        input_u64,
        input_i8,
        input_i16,
        input_i32,
        input_i64,
        input_float,
        input_double,
        input_half,
        input_2d,
        input_3d,
        output_u8,
        output_u16,
        output_u32,
        output_u64,
        output_i8,
        output_i16,
        output_i32,
        output_i64,
        output_float,
        output_double,
        output_half,
        output_2d,
        output_3d,
    )

    combinations = [
        ("u8", input_u8, output_u8, scalar_u8),
        ("u16", input_u16, output_u16, scalar_u16),
        ("u32", input_u32, output_u32, scalar_u32),
        ("u64", input_u64, output_u64, scalar_u64),
        ("i8", input_i8, output_i8, scalar_i8),
        ("i16", input_i16, output_i16, scalar_i16),
        ("i32", input_i32, output_i32, scalar_i32),
        ("i64", input_i64, output_i64, scalar_i64),
        ("float", input_float, output_float, scalar_float),
        ("double", input_double, output_double, scalar_double),
        ("half", input_half, output_half, scalar_float),
    ]

    for _, input, output, scalar in combinations:
        for i, o in zip(input, output):
            scalar_as_numpy = numpy.array(scalar).astype(input.dtype)
            assert abs(o - (i + scalar_as_numpy)) < ERROR_THRESHOLD

    for x in range(input_2d.shape[0]):
        for y in range(input_2d.shape[1]):
            assert output_2d[x, y] == input_2d[x, y] + scalar_i8

    for x in range(input_3d.shape[0]):
        for y in range(input_3d.shape[1]):
            for z in range(input_3d.shape[2]):
                assert output_3d[x, y, z] == input_3d[x, y, z] + scalar_i8 + offset

    try:
        # Expected requirement failure #1
        scalar_i32 = 0
        addconstant_impl_func(
            scalar_u1,
            scalar_u8,
            scalar_u16,
            scalar_u32,
            scalar_u64,
            scalar_i8,
            scalar_i16,
            scalar_i32,
            scalar_i64,
            scalar_float,
            scalar_double,
            input_u8,
            input_u16,
            input_u32,
            input_u64,
            input_i8,
            input_i16,
            input_i32,
            input_i64,
            input_float,
            input_double,
            input_half,
            input_2d,
            input_3d,
            output_u8,
            output_u16,
            output_u32,
            output_u64,
            output_i8,
            output_i16,
            output_i32,
            output_i64,
            output_float,
            output_double,
            output_half,
            output_2d,
            output_3d,
        )
    except RuntimeError as e:
        assert str(e) == "Halide Runtime Error: -27", e
    else:
        assert False, "Did not see expected exception!"

    try:
        # Expected requirement failure #2 -- note that for AOT-compiled
        # code in Python, the error message is stricly numeric (the text
        # of the error isn't currently propagated int he exception).
        scalar_i32 = -1
        addconstant_impl_func(
            scalar_u1,
            scalar_u8,
            scalar_u16,
            scalar_u32,
            scalar_u64,
            scalar_i8,
            scalar_i16,
            scalar_i32,
            scalar_i64,
            scalar_float,
            scalar_double,
            input_u8,
            input_u16,
            input_u32,
            input_u64,
            input_i8,
            input_i16,
            input_i32,
            input_i64,
            input_float,
            input_double,
            input_half,
            input_2d,
            input_3d,
            output_u8,
            output_u16,
            output_u32,
            output_u64,
            output_i8,
            output_i16,
            output_i32,
            output_i64,
            output_float,
            output_double,
            output_half,
            output_2d,
            output_3d,
        )
    except RuntimeError as e:
        assert str(e) == "Halide Runtime Error: -27", e
    else:
        assert False, "Did not see expected exception!"


if __name__ == "__main__":
    for t, o in TESTS_AND_OFFSETS:
        test(t, o)



"""Synthesizes the cpp wrapper code and builds dynamic Python extension."""
import os
import platform
import re
from setuptools import setup, find_packages

from torch.utils.cpp_extension import BuildExtension
import torch as th


def generate_pybind_wrapper(path, headers, has_cuda):
    s = "#include \"torch/extension.h\"\n\n"
    if has_cuda:
        s += "#include \"HalidePyTorchCudaHelpers.h\"\n"
    s += "#include \"HalidePyTorchHelpers.h\"\n"
    for h in headers:
        s += "#include \"{}\"\n".format(os.path.splitext(h)[0]+".pytorch.h")

    s += "\nPYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {\n"
    for h in headers:
        name = os.path.splitext(h)[0]
        s += "  m.def(\"{}\", &{}_th_, \"PyTorch wrapper of the Halide pipeline {}\");\n".format(
          name, name, name)
    s += "}\n"
    with open(path, 'w') as fid:
        fid.write(s)

if __name__ == "__main__":
    # This is where the generate Halide ops headers live. We also generate the .cpp
    # wrapper in this directory
    build_dir = os.getenv("BIN")
    if build_dir is None or not os.path.exists(build_dir):
        raise ValueError("Bin directory {} is invalid".format(build_dir))

    # Path to a distribution of Halide
    halide_dir = os.getenv("HALIDE_DISTRIB_PATH")
    if halide_dir is None or not os.path.exists(halide_dir):
        raise ValueError("Halide directory {} is invalid".format(halide_dir))

    has_cuda = os.getenv("HAS_CUDA")
    if has_cuda is None or has_cuda == "0":
        has_cuda = False
    else:
        has_cuda = True

    include_dirs = [build_dir, os.path.join(halide_dir, "include")]
    # Note that recent versions of PyTorch (at least 1.7.1) requires C++14
    # in order to compile extensions
    compile_args = ["-std=c++14", "-g"]
    if platform.system() == "Darwin":  # on osx libstdc++ causes trouble
        compile_args += ["-stdlib=libc++"]

    re_cc = re.compile(r".*\.pytorch\.h")
    hl_srcs = [f for f in os.listdir(build_dir) if re_cc.match(f)]

    ext_name = "halide_ops"
    hl_libs = []  # Halide op libraries to link to
    hl_headers = []  # Halide op headers to include in the wrapper
    for f in hl_srcs:
        # Add all Halide generated torch wrapper
        hl_src = os.path.join(build_dir, f)

        # Add all Halide-generated libraries
        hl_lib = hl_src.split(".")[0] + ".a"
        hl_libs.append(hl_lib)

        hl_header = hl_src.split(".")[0] + ".h"
        hl_headers.append(os.path.basename(hl_header))

    # C++ wrapper code that includes so that we get all the Halide ops in a
    # single python extension
    wrapper_path = os.path.join(build_dir, "pybind_wrapper.cpp")
    sources = [wrapper_path]

    if has_cuda:
        print("Generating CUDA wrapper")
        generate_pybind_wrapper(wrapper_path, hl_headers, True)
        from torch.utils.cpp_extension import CUDAExtension
        extension = CUDAExtension(ext_name, sources,
                                  include_dirs=include_dirs,
                                  extra_objects=hl_libs,
                                  libraries=["cuda"],  # Halide ops need the full cuda lib, not just the RT library
                                  extra_compile_args=compile_args)
    else:
        print("Generating CPU wrapper")
        generate_pybind_wrapper(wrapper_path, hl_headers, False)
        from torch.utils.cpp_extension import CppExtension
        extension = CppExtension(ext_name, sources,
                                 include_dirs=include_dirs,
                                 extra_objects=hl_libs,
                                 extra_compile_args=compile_args)

    # Build the Python extension module
    setup(name=ext_name,
          verbose=True,
          url="",
          author_email="your@email.com",
          author="Some Author",
          version="0.0.0",
          ext_modules=[extension],
          cmdclass={"build_ext": BuildExtension}
          )



import halide as hl
import numpy as np

from simplepy_generator import SimplePy
import simplecpp_pystub  # Needed for create_callable_from_generator("simplecpp") to work


def test_callable():
    p_int16 = hl.Param(hl.Int(16), 42)
    p_float = hl.Param(hl.Float(32), 1.0)
    p_img = hl.ImageParam(hl.UInt(8), 2)

    x = hl.Var("x")
    y = hl.Var("y")
    f = hl.Func("f")

    f[x, y] = p_img[x, y] + hl.u8(p_int16 / p_float)

    in1 = hl.Buffer(hl.UInt(8), [10, 10])
    in2 = hl.Buffer(hl.UInt(8), [10, 10])

    for i in range(10):
        for j in range(10):
            in1[i, j] = i + j * 10
            in2[i, j] = i * 10 + j

    c = f.compile_to_callable([p_img, p_int16, p_float])

    out1 = hl.Buffer(hl.UInt(8), [10, 10])
    c(in1, 42, 1.0, out1)

    out2 = hl.Buffer(hl.UInt(8), [10, 10])
    c(in2, 22, 2.0, out2)

    out3 = hl.Buffer(hl.UInt(8), [10, 10])
    c(in1, 12, 1.0, out3)

    out4 = hl.Buffer(hl.UInt(8), [10, 10])
    c(in2, 16, 1.0, out4)

    for i in range(10):
        for j in range(10):
            assert out1[i, j] == i + j * 10 + 42
            assert out2[i, j] == i * 10 + j + 11
            assert out3[i, j] == i + j * 10 + 12
            assert out4[i, j] == i * 10 + j + 16

    # Test bounds inference. Note that in Python there
    # isn't a "natural" way to create a buffer with a null host ptr
    # so we use this specific API for the purpose.
    in_bounds = hl.Buffer.make_bounds_query(hl.UInt(8), [1, 1])
    out_bounds = hl.Buffer.make_bounds_query(hl.UInt(8), [20, 20])
    c(in_bounds, 42, 1.0, out_bounds)

    assert in_bounds.defined()
    assert in_bounds.dim(0).extent() == 20
    assert in_bounds.dim(1).extent() == 20
    assert in1.dim(0).extent() == 10
    assert in1.dim(1).extent() == 10


def test_simple(callable_factory):
    x, y = hl.Var(), hl.Var()
    target = hl.get_jit_target_from_environment()

    b_in = hl.Buffer(hl.UInt(8), [2, 2])
    b_in.fill(123)

    float_in = 3.5

    b_out = hl.Buffer(hl.Float(32), [2, 2])

    def _check(offset=0):
        assert b_out[0, 0] == float_in + offset + 123
        assert b_out[0, 1] == float_in + offset + 123
        assert b_out[1, 0] == float_in + offset + 123
        assert b_out[1, 1] == float_in + offset + 123

    generator_params = {}
    simple = callable_factory(target, generator_params)

    # ----------- Positional arguments
    simple(b_in, float_in, b_out)
    _check()

    # ----------- Keyword arguments
    # Natural order
    simple(buffer_input=b_in, float_arg=float_in, simple_output=b_out)
    _check()

    # Weird order
    simple(float_arg=float_in, simple_output=b_out, buffer_input=b_in)
    _check()

    # ----------- Positional + Keywords

    # Natural order
    simple(b_in, simple_output=b_out, float_arg=float_in)
    _check()

    # Weird order
    simple(b_in, float_in, simple_output=b_out)
    _check()

    # ----------- Above set again, w/ additional GeneratorParam mixed in
    k = 42

    generator_params = {"offset": str(k)}
    simple_42 = callable_factory(target, generator_params)
    simple_42(b_in, float_in, b_out)
    _check(k)

    # ----------- Test various failure modes
    try:
        # too many positional args
        simple(b_in, float_in, 4, b_out)
    except hl.HalideError as e:
        assert "Expected at most 3 positional arguments, but saw 4." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # too few positional args
        simple(b_in)
    except hl.HalideError as e:
        assert "Expected exactly 3 positional arguments, but saw 1." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Inputs that can't be converted to what the receiver needs (positional)
        simple(hl.f32(3.141592), "happy", b_out)
    except hl.HalideError as e:
        assert "is not an instance of" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Inputs that can't be converted to what the receiver needs (named)
        simple(b_in, float_in, simple_output="bogus")
    except hl.HalideError as e:
        assert "is not an instance of" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Bad keyword argument
        simple(buffer_input=b_in, float_arg=float_in, funk_input=b_out)
    except hl.HalideError as e:
        assert "Unknown argument 'funk_input' specified via keyword." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # too few keyword args
        simple(float_arg=float_in, simple_output=b_out)
    except hl.HalideError as e:
        assert (
            "Argument buffer_input was not specified by either positional or keyword argument."
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"

    try:
        # Arg specified by pos + kw
        simple(b_in, buffer_input=b_in, float_arg=float_in, simple_output=b_out)
    except hl.HalideError as e:
        assert "Argument buffer_input specified multiple times." in str(e)
    else:
        assert False, "Did not see expected exception!"


if __name__ == "__main__":
    # test_callable()

    def via_simplecpp_pystub(target, generator_params):
        return hl.create_callable_from_generator(target, "simplecpp", generator_params)

    def via_simplepy(target, generator_params):
        with hl.GeneratorContext(target):
            g = SimplePy(generator_params=generator_params)
            return g.compile_to_callable()

    test_simple(via_simplecpp_pystub)
    test_simple(via_simplepy)



import halide as hl
import numpy as np


def test_compiletime_error():
    x = hl.Var("x")
    y = hl.Var("y")
    f = hl.Func("f")
    f[x, y] = hl.u16(x + y)
    # Deliberate type-mismatch error
    buf = hl.Buffer(hl.UInt(8), [2, 2])
    try:
        f.realize(buf)
    except hl.HalideError as e:
        assert (
            "Output buffer f has type uint16 but type of the buffer passed in is uint8"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"


def test_runtime_error():
    x = hl.Var("x")
    f = hl.Func("f")
    f[x] = hl.u8(x)
    f.bound(x, 0, 1)
    # Deliberate runtime error
    buf = hl.Buffer(hl.UInt(8), [10])
    try:
        f.realize(buf)
    except hl.HalideError as e:
        assert "do not cover required region" in str(e)
    else:
        assert False, "Did not see expected exception!"


def test_misused_and():
    x = hl.Var("x")
    y = hl.Var("y")
    f = hl.Func("f")
    try:
        f[x, y] = hl.print_when(x == 0 and y == 0, 0, "x=", x, "y=", y)
        f.realize([10, 10])
    except ValueError as e:
        assert "cannot be converted to a bool" in str(e)
    else:
        assert False, "Did not see expected exception!"


def test_misused_or():
    x = hl.Var("x")
    y = hl.Var("y")
    f = hl.Func("f")
    try:
        f[x, y] = hl.print_when(x == 0 or y == 0, 0, "x=", x, "y=", y)
        f.realize([10, 10])
    except ValueError as e:
        assert "cannot be converted to a bool" in str(e)
    else:
        assert False, "Did not see expected exception!"


def test_basics():
    input = hl.ImageParam(hl.UInt(16), 2, "input")
    x, y = hl.Var("x"), hl.Var("y")

    blur_x = hl.Func("blur_x")
    blur_xx = hl.Func("blur_xx")
    blur_y = hl.Func("blur_y")

    yy = hl.i32(1)
    assert yy.type() == hl.Int(32)

    z = x + 1
    input[x, y]
    input[0, 0]
    input[z, y]
    input[x + 1, y]
    input[x, y] + input[x + 1, y]

    if False:
        aa = blur_x[x, y]
        bb = blur_x[x, y + 1]
        aa + bb
        blur_x[x, y] + blur_x[x, y + 1]

    (input[x, y] + input[x + 1, y]) / 2
    blur_x[x, y]
    blur_xx[x, y] = input[x, y]

    blur_x[x, y] = (input[x, y] + input[x + 1, y] + input[x + 2, y]) / 3
    blur_y[x, y] = (blur_x[x, y] + blur_x[x, y + 1] + blur_x[x, y + 2]) / 3

    xi, yi = hl.Var("xi"), hl.Var("yi")
    blur_y.tile(x, y, xi, yi, 8, 4).parallel(y).vectorize(xi, 8)
    blur_x.compute_at(blur_y, x).vectorize(x, 8)
    blur_y.compile_jit()


def test_basics2():
    input = hl.ImageParam(hl.Float(32), 3, "input")
    r_sigma = hl.Param(hl.Float(32), "r_sigma", 0.1)
    s_sigma = 8

    x = hl.Var("x")
    y = hl.Var("y")
    z = hl.Var("z")
    c = hl.Var("c")

    # Add a boundary condition
    clamped = hl.Func("clamped")
    clamped[x, y] = input[
        hl.clamp(x, 0, input.width() - 1), hl.clamp(y, 0, input.height() - 1), 0
    ]

    # Construct the bilateral grid
    r = hl.RDom([(0, s_sigma), (0, s_sigma)], "r")
    val0 = clamped[x * s_sigma, y * s_sigma]
    val00 = clamped[x * s_sigma * hl.i32(1), y * s_sigma * hl.i32(1)]
    val22 = clamped[
        x * s_sigma - hl.i32(s_sigma // 2), y * s_sigma - hl.i32(s_sigma // 2)
    ]
    val2 = clamped[x * s_sigma - s_sigma // 2, y * s_sigma - s_sigma // 2]
    val3 = clamped[x * s_sigma + r.x - s_sigma // 2, y * s_sigma + r.y - s_sigma // 2]

    try:
        val1 = clamped[x * s_sigma - s_sigma / 2, y * s_sigma - s_sigma / 2]
    except hl.HalideError as e:
        assert "Implicit cast from float32 to int" in str(e)
    else:
        assert False, "Did not see expected exception!"


def test_basics3():
    input = hl.ImageParam(hl.Float(32), 3, "input")
    r_sigma = hl.Param(
        hl.Float(32), "r_sigma", 0.1
    )  # Value needed if not generating an executable
    s_sigma = 8  # This is passed during code generation in the C++ version

    x = hl.Var("x")
    y = hl.Var("y")
    z = hl.Var("z")
    c = hl.Var("c")

    # Add a boundary condition
    clamped = hl.Func("clamped")
    clamped[x, y] = input[
        hl.clamp(x, 0, input.width() - 1), hl.clamp(y, 0, input.height() - 1), 0
    ]

    # Construct the bilateral grid
    r = hl.RDom([(0, s_sigma), (0, s_sigma)], "r")
    val = clamped[x * s_sigma + r.x - s_sigma // 2, y * s_sigma + r.y - s_sigma // 2]
    val = hl.clamp(val, 0.0, 1.0)
    zi = hl.i32((val / r_sigma) + 0.5)
    histogram = hl.Func("histogram")
    histogram[x, y, z, c] = 0.0

    ss = hl.select(c == 0, val, 1.0)
    left = histogram[x, y, zi, c]
    left += 5
    left += ss


def test_basics4():
    # Test for f[g[r]] = ...
    # See https://github.com/halide/Halide/issues/4285
    x = hl.Var("x")
    f = hl.Func("f")
    g = hl.Func("g")
    g[x] = 1
    f[x] = 0.0
    r = hl.RDom([(0, 100)])
    f[g[r]] = 2.5
    f.compute_root()
    f.compile_jit()


def test_basics5():
    # Test Func.in_()
    x, y = hl.Var("x"), hl.Var("y")
    f = hl.Func("f")
    g = hl.Func("g")
    h = hl.Func("h")
    f[x, y] = y
    r = hl.RDom([(0, 100)])
    g[x] = 0
    g[x] += f[x, r]
    h[x] = 0
    h[x] += f[x, r]
    f.in_(g).compute_at(g, x)
    f.in_(h).compute_at(h, x)
    g.compute_root()
    h.compute_root()
    p = hl.Pipeline([g, h])
    p.compile_jit()


def test_float_or_int():
    x = hl.Var("x")
    i32, f32 = hl.Int(32), hl.Float(32)

    assert hl.Expr(x).type() == i32
    assert (x * 2).type() == i32
    assert (x / 2).type() == i32
    assert ((x // 2) - 1 + 2 * (x % 2)).type() == i32
    assert ((x / 2) - 1 + 2 * (x % 2)).type() == i32
    assert ((x / 2)).type() == i32
    assert ((x / 2.0)).type() == f32
    assert ((x // 2)).type() == i32
    assert ((x // 2) - 1).type() == i32
    assert ((x % 2)).type() == i32
    assert (2 * (x % 2)).type() == i32
    assert ((x // 2) - 1 + 2 * (x % 2)).type() == i32

    assert type(x) == hl.Var
    assert (hl.Expr(x)).type() == i32
    assert (hl.Expr(2.0)).type() == f32
    assert (hl.Expr(2)).type() == i32
    assert (x + 2).type() == i32
    assert (2 + x).type() == i32
    assert (hl.Expr(2) + hl.Expr(3)).type() == i32
    assert (hl.Expr(2.0) + hl.Expr(3)).type() == f32
    assert (hl.Expr(2) + 3.0).type() == f32
    assert (hl.Expr(2) + 3).type() == i32
    assert (hl.Expr(x) + 2).type() == i32
    assert (2 + hl.Expr(x)).type() == i32
    assert (2 * (x + 2)).type() == i32
    assert (x + 0).type() == i32
    assert (x % 2).type() == i32
    assert (2 * x).type() == i32
    assert (x * 2).type() == i32
    assert (x * 2).type() == i32
    assert ((x % 2)).type() == i32
    assert ((x % 2) * 2).type() == i32
    assert (2 * (x % 2)).type() == i32
    assert ((x + 2) * 2).type() == i32


def test_operator_order():
    x = hl.Var("x")
    f = hl.Func("f")
    x + 1
    1 + x
    f[x] = x**2
    f[x] + 1
    hl.Expr(1) + f[x]
    1 + f[x]


def _check_is_u16(e):
    assert e.type() == hl.UInt(16), e.type()


def test_int_promotion():
    # Verify that (Exprlike op literal) correctly matches the type
    # of the literal to the Exprlike (rather than promoting the result to int32).
    # All types that use add_binary_operators() should be tested here.

    x = hl.Var("x")
    # All the binary ops are handled the same, so + is good enough

    # Exprlike = FuncRef
    f = hl.Func("f")
    f[x] = hl.u16(x)
    _check_is_u16(f[x] + 2)
    _check_is_u16(2 + f[x])

    # Exprlike = Expr
    e = hl.Expr(f[x])
    _check_is_u16(e + 2)
    _check_is_u16(2 + e)

    # Exprlike = Param
    p = hl.Param(hl.UInt(16))
    _check_is_u16(p + 2)
    _check_is_u16(2 + p)

    # Exprlike = RDom/RVar
    # Exprlike = Var
    # (skipped, since these can never have values of any type other than int32)


def test_vector_tile():
    # Test Func.tile() and Stage.tile() with vector arguments
    x, y, z = [hl.Var(c) for c in "xyz"]
    xi, yi, zi = [hl.Var(c + "i") for c in "xyz"]
    xo, yo, zo = [hl.Var(c + "o") for c in "xyz"]
    f = hl.Func("f")
    g = hl.Func("g")
    h = hl.Func("h")
    f[x, y] = y
    f[x, y] += x
    g[x, y, z] = x + y
    g[x, y, z] += z
    f.tile([x, y], [xo, yo], [x, y], [8, 8])
    f.update(0).tile([x, y], [xo, yo], [xi, yi], [8, 8])
    g.tile([x, y], [xo, yo], [x, y], [8, 8], hl.TailStrategy.RoundUp)
    g.update(0).tile([x, y], [xo, yo], [xi, yi], [8, 8], hl.TailStrategy.GuardWithIf)
    p = hl.Pipeline([f, g])
    p.compile_jit()


def test_scalar_funcs():
    input = hl.ImageParam(hl.UInt(16), 0, "input")

    f = hl.Func("f")
    g = hl.Func("g")

    input[()]

    (input[()] + input[()]) / 2
    f[()]
    g[()]

    f[()] = (input[()] + input[()] + input[()]) / 3
    g[()] = (f[()] + f[()] + f[()]) / 3

    g.compile_jit()


def test_bool_conversion():
    x = hl.Var("x")
    f = hl.Func("f")
    f[x] = x
    s = bool(True)
    # Verify that this doesn't fail with 'Argument passed to specialize must be of type bool'
    f.compute_root().specialize(True)


def test_typed_funcs():
    x = hl.Var("x")
    y = hl.Var("y")

    f = hl.Func("f")
    assert not f.defined()
    try:
        assert f.type() == Int(32)
    except hl.HalideError as e:
        assert "it is undefined" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        assert f.outputs() == 0
    except hl.HalideError as e:
        assert "it is undefined" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        assert f.dimensions() == 0
    except hl.HalideError as e:
        assert "it is undefined" in str(e)
    else:
        assert False, "Did not see expected exception!"

    f = hl.Func(hl.Int(32), 2, "f")
    assert not f.defined()
    assert f.type() == hl.Int(32)
    assert f.types() == [hl.Int(32)]
    assert f.outputs() == 1
    assert f.dimensions() == 2

    f = hl.Func([hl.Int(32), hl.Float(64)], 3, "f")
    assert not f.defined()
    try:
        assert f.type() == hl.Int(32)
    except hl.HalideError as e:
        assert "it returns a Tuple" in str(e)
    else:
        assert False, "Did not see expected exception!"

    assert f.types() == [hl.Int(32), hl.Float(64)]
    assert f.outputs() == 2
    assert f.dimensions() == 3

    f = hl.Func(hl.Int(32), 1, "f")
    try:
        f[x, y] = hl.i32(0)
        f.realize([10, 10])
    except hl.HalideError as e:
        assert (
            "is constrained to have exactly 1 dimensions, but is defined with 2 dimensions"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"

    f = hl.Func(hl.Int(32), 2, "f")
    try:
        f[x, y] = hl.i16(0)
        f.realize([10, 10])
    except hl.HalideError as e:
        assert (
            "is constrained to only hold values of type int32 but is defined with values of type int16"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"

    f = hl.Func((hl.Int(32), hl.Float(32)), 2, "f")
    try:
        f[x, y] = (hl.i16(0), hl.f64(0))
        f.realize([10, 10])
    except hl.HalideError as e:
        assert (
            "is constrained to only hold values of type (int32, float32) but is defined with values of type (int16, float64)"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"


def test_requirements():
    delta = hl.Param(hl.Int(32), "delta")
    x = hl.Var("x")
    f = hl.Func("f_requirements")
    f[x] = x + delta

    # Add a requirement
    p = hl.Pipeline([f])
    p.add_requirement(delta != 0)  # error_args omitted
    p.add_requirement(delta > 0, "negative values are bad", delta)

    delta.set(1)
    p.realize([10])

    try:
        delta.set(0)
        p.realize([10])
    except hl.HalideError as e:
        assert "Requirement Failed: (false)" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        delta.set(-1)
        p.realize([10])
    except hl.HalideError as e:
        assert "Requirement Failed: (false) negative values are bad -1" in str(e)
    else:
        assert False, "Did not see expected exception!"


if __name__ == "__main__":
    test_compiletime_error()
    test_runtime_error()
    test_misused_and()
    test_misused_or()
    test_typed_funcs()
    test_float_or_int()
    test_operator_order()
    test_int_promotion()
    test_vector_tile()
    test_basics()
    test_basics2()
    test_basics3()
    test_basics4()
    test_basics5()
    test_scalar_funcs()
    test_bool_conversion()
    test_requirements()



import os, shutil, tempfile
import halide as hl


def main():
    x = hl.Var("x")

    f = hl.Func("f")
    f[x] = 100 * x

    args = []

    tmpdir = tempfile.mkdtemp()
    try:
        p = os.path.join(tmpdir, "f.bc")
        f.compile_to_bitcode(p, args, "f")
        assert os.path.isfile(p)

        p = os.path.join(tmpdir, "f.cpp")
        f.compile_to_c(p, args, "f")
        assert os.path.isfile(p)

        p = os.path.join(tmpdir, "f.o")
        f.compile_to_object(p, args, "f")
        assert os.path.isfile(p)

        p = os.path.join(tmpdir, "f.h")
        f.compile_to_header(p, args, "f")
        assert os.path.isfile(p)

        p = os.path.join(tmpdir, "f.s")
        f.compile_to_assembly(p, args, "f")
        assert os.path.isfile(p)

        p = os.path.join(tmpdir, "f.txt")
        f.compile_to_lowered_stmt(p, args)
        assert os.path.isfile(p)

        f.compile_to_file(os.path.join(tmpdir, "f_all"), args)
        assert os.path.isfile(os.path.join(tmpdir, "f_all.h"))
        if hl.get_target_from_environment().os == hl.TargetOS.Windows:
            assert os.path.isfile(os.path.join(tmpdir, "f_all.obj"))
        else:
            assert os.path.isfile(os.path.join(tmpdir, "f_all.o"))

        p = os.path.join(tmpdir, "f.stmt.html")
        f.compile_to({hl.OutputFileType.stmt_html: p}, args, "f")
        assert os.path.isfile(p)

    finally:
        shutil.rmtree(tmpdir, ignore_errors=True)


if __name__ == "__main__":
    main()



import halide as hl
import io
import contextlib


def test_warnings():
    i = hl.Var()
    f = hl.Func("f")
    f[i] = 0.0
    f.bound(i, 0, 127)

    g = hl.Func("g")
    g[i] = f[i * i]

    expected_warning = (
        "Warning: It is meaningless to bound dimension v0 of function f to be within [0, 127] because "
        "the function is scheduled inline.\n"
    )

    buffer = io.StringIO()
    with contextlib.redirect_stdout(buffer):
        g.realize([16])

    buffer.seek(0)
    stdout_lines = buffer.readlines()
    assert len(stdout_lines) > 0
    for line in stdout_lines:
        assert line == expected_warning

def test_unscheduled(suppress):
    x = hl.Var()
    f = hl.Func("f_%s" % str(suppress))
    f[x] = 0
    f[x] += 5
    f.vectorize(x, 8)
    if suppress:
        f.update(0).unscheduled()

    buffer = io.StringIO()
    with contextlib.redirect_stdout(buffer):
        f.realize([1024])

    buffer.seek(0)
    stdout_lines = buffer.readlines()
    if suppress:
        assert len(stdout_lines) == 0
    else:
        expected_warning = "Warning: Update definition 0 of function f_False has not been scheduled"
        assert len(stdout_lines) > 0
        for line in stdout_lines:
            assert line.startswith(expected_warning), "\n%s\n%s" % (line, expected_warning)

if __name__ == "__main__":
    test_warnings()
    test_unscheduled(True)
    test_unscheduled(False)



import halide as hl


def test_rdom():
    x = hl.Var("x")
    y = hl.Var("y")

    diagonal = hl.Func("diagonal")
    diagonal[x, y] = 1

    domain_width = 10
    domain_height = 10

    r = hl.RDom([(0, domain_width), (0, domain_height)])
    r.where(r.x <= r.y)

    diagonal[r.x, r.y] += 2
    output = diagonal.realize([domain_width, domain_height])

    for iy in range(domain_height):
        for ix in range(domain_width):
            if ix <= iy:
                assert output[ix, iy] == 3
            else:
                assert output[ix, iy] == 1

    assert r.x.name() == r[0].name()
    assert r.y.name() == r[1].name()
    try:
        r[-1].name()
        raise Exception("underflowing index should raise KeyError")
    except KeyError:
        pass
    try:
        r[2].name()
        raise Exception("overflowing index should raise KeyError")
    except KeyError:
        pass
    try:
        r["foo"].name()
        raise Exception("bad index type should raise TypeError")
    except TypeError:
        pass

    return 0


if __name__ == "__main__":
    test_rdom()



import contextlib
import halide as hl
import sys
import io

# redirect_stdout() requires Python3, alas
@contextlib.contextmanager
def _redirect_stdout(out):
    old_out = sys.stdout
    sys.stdout = out
    try:
        yield out
    finally:
        sys.stdout = old_out


def test_print_expr():
    x = hl.Var("x")
    f = hl.Func("f")
    f[x] = hl.print(
        hl.cast(hl.UInt(8), x), "is what", "the", 1, "and", hl.f32(3.1415), "saw"
    )
    buf = hl.Buffer(hl.UInt(8), [1])
    output = io.StringIO()
    with _redirect_stdout(output):
        f.realize(buf)
        expected = "0 is what the 1 and 3.141500 saw\n"
        actual = output.getvalue()
        assert expected == actual, "Expected: %s, Actual: %s" % (expected, actual)


def test_print_when():
    x = hl.Var("x")
    f = hl.Func("f")
    f[x] = hl.print_when(x == 3, hl.cast(hl.UInt(8), x * x), "is result at", x)
    buf = hl.Buffer(hl.UInt(8), [10])
    output = io.StringIO()
    with _redirect_stdout(output):
        f.realize(buf)
        expected = "9 is result at 3\n"
        actual = output.getvalue()
        assert expected == actual, "Expected: %s, Actual: %s" % (expected, actual)


def test_select():
    x = hl.Var()
    f = hl.Func()
    # fmt: off
    f[x] = hl.select(x == 0, 31,
                     x == 2, (x * 24),
                     x == 2, 999,  # should be ignored: first condition wins
                             x)
    # fmt: on
    b = f.realize([4])
    assert b[0] == 31
    assert b[1] == 1
    assert b[2] == 48
    assert b[3] == 3


def test_mux():
    c = hl.Var()
    f = hl.Func()
    f[c] = hl.mux(c, [123, 456, c])
    b = f.realize([4])
    assert b[0] == 123
    assert b[1] == 456
    assert b[2] == 2
    assert b[3] == 3


def test_mux_tuple():
    f = hl.Func()
    g = hl.Func()
    x = hl.Var()
    c = hl.Var()
    g[x] = (123, 456, x)
    f[x, c] = hl.mux(c, g[x])
    b = f.realize([1, 4])
    assert b[0, 0] == 123
    assert b[0, 1] == 456
    assert b[0, 2] == 0
    assert b[0, 3] == 0


def test_minmax():
    x = hl.Var()
    f = hl.Func()
    # fmt: off
    f[x] = hl.select(x == 0,              hl.min(x, 1),
                     (x == 2) | (x == 4), hl.i32(hl.min(hl.f32(x), hl.f32(3.2), x*hl.f32(2.1))),
                     x == 3,              hl.max(x, x * 3, 1, x * 4),
                                          x)
    # fmt: on
    b = f.realize([5])
    assert b[0] == 0
    assert b[1] == 1, b[1]
    assert b[2] == 2
    assert b[3] == 12
    assert b[4] == 3


if __name__ == "__main__":
    test_print_expr()
    test_print_when()
    test_select()
    test_mux()
    test_minmax()



"""
Bilateral histogram.
"""

import halide as hl


@hl.alias(
    bilateral_grid_Adams2019={"autoscheduler": "Adams2019"},
    bilateral_grid_Mullapudi2016={"autoscheduler": "Mullapudi2016"},
    bilateral_grid_Li2018={"autoscheduler": "Li2018"},
)
@hl.generator()
class bilateral_grid:
    s_sigma = hl.GeneratorParam(8)

    input_buf = hl.InputBuffer(hl.Float(32), 2)
    r_sigma = hl.InputScalar(hl.Float(32))
    bilateral_grid = hl.OutputBuffer(hl.Float(32), 2)

    def generate(self):
        g = self

        x, y, z, c = hl.vars("x y z c")

        # Add a boundary condition
        clamped = hl.BoundaryConditions.repeat_edge(g.input_buf)

        # Construct the bilateral grid
        r = hl.RDom([(0, g.s_sigma), (0, g.s_sigma)])
        val = clamped[
            x * g.s_sigma + r.x - g.s_sigma // 2,
            y * g.s_sigma + r.y - g.s_sigma // 2,
        ]
        val = hl.clamp(val, 0.0, 1.0)

        zi = hl.i32(val / g.r_sigma + 0.5)

        histogram = hl.Func("histogram")
        histogram[x, y, z, c] = 0.0
        histogram[x, y, zi, c] += hl.mux(c, [val, 1.0])

        # Blur the histogram using a five-tap filter
        blurx, blury, blurz = hl.funcs("blurx blury blurz")
        blurz[x, y, z, c] = (
            histogram[x, y, z - 2, c]
            + histogram[x, y, z - 1, c] * 4
            + histogram[x, y, z, c] * 6
            + histogram[x, y, z + 1, c] * 4
            + histogram[x, y, z + 2, c]
        )
        blurx[x, y, z, c] = (
            blurz[x - 2, y, z, c]
            + blurz[x - 1, y, z, c] * 4
            + blurz[x, y, z, c] * 6
            + blurz[x + 1, y, z, c] * 4
            + blurz[x + 2, y, z, c]
        )
        blury[x, y, z, c] = (
            blurx[x, y - 2, z, c]
            + blurx[x, y - 1, z, c] * 4
            + blurx[x, y, z, c] * 6
            + blurx[x, y + 1, z, c] * 4
            + blurx[x, y + 2, z, c]
        )

        # Take trilinear samples to compute the output
        val = hl.clamp(clamped[x, y], 0.0, 1.0)
        zv = val / g.r_sigma
        zi = hl.i32(zv)
        zf = zv - zi
        xf = hl.f32(x % g.s_sigma) / g.s_sigma
        yf = hl.f32(y % g.s_sigma) / g.s_sigma
        xi = x / g.s_sigma
        yi = y / g.s_sigma

        interpolated = hl.Func("interpolated")
        interpolated[x, y, c] = hl.lerp(
            hl.lerp(
                hl.lerp(blury[xi, yi, zi, c], blury[xi + 1, yi, zi, c], xf),
                hl.lerp(blury[xi, yi + 1, zi, c], blury[xi + 1, yi + 1, zi, c], xf),
                yf,
            ),
            hl.lerp(
                hl.lerp(blury[xi, yi, zi + 1, c], blury[xi + 1, yi, zi + 1, c], xf),
                hl.lerp(
                    blury[xi, yi + 1, zi + 1, c], blury[xi + 1, yi + 1, zi + 1, c], xf
                ),
                yf,
            ),
            zf,
        )

        # Normalize
        g.bilateral_grid[x, y] = interpolated[x, y, 0] / interpolated[x, y, 1]

        # ESTIMATES
        # (This can be useful in conjunction with RunGen and benchmarks as well
        # as auto-schedule, so we do it in all cases.)
        # Provide estimates on the input image
        g.input_buf.set_estimates([(0, 1536), (0, 2560)])
        # Provide estimates on the parameters
        g.r_sigma.set_estimate(0.1)
        # TODO: Compute estimates from the parameter values
        histogram.set_estimate(z, -2, 16)
        blurz.set_estimate(z, 0, 12)
        blurx.set_estimate(z, 0, 12)
        blury.set_estimate(z, 0, 12)
        g.bilateral_grid.set_estimates([(0, 1536), (0, 2560)])

        if g.using_autoscheduler():
            # nothing
            pass
        elif g.target().has_gpu_feature():
            # 0.50ms on an RTX 2060

            xi, yi, zi = hl.vars("xi yi zi")

            # Schedule blurz in 8x8 tiles. This is a tile in
            # grid-space, which means it represents something like
            # 64x64 pixels in the input (if s_sigma is 8).
            blurz.compute_root().reorder(c, z, x, y).gpu_tile(x, y, xi, yi, 8, 8)

            # Schedule histogram to happen per-tile of blurz, with
            # intermediate results in shared memory. This means histogram
            # and blurz makes a three-stage kernel:
            # 1) Zero out the 8x8 set of histograms
            # 2) Compute those histogram by iterating over lots of the input image
            # 3) Blur the set of histograms in z
            histogram.reorder(c, z, x, y).compute_at(blurz, x).gpu_threads(x, y)
            histogram.update().reorder(c, r.x, r.y, x, y).gpu_threads(x, y).unroll(c)

            # Schedule the remaining blurs and the sampling at the end
            # similarly.
            (
                blurx.compute_root()
                .reorder(c, x, y, z)
                .reorder_storage(c, x, y, z)
                .vectorize(c)
                .unroll(y, 2, hl.TailStrategy.RoundUp)
                .gpu_tile(x, y, z, xi, yi, zi, 32, 8, 1, hl.TailStrategy.RoundUp)
            )
            (
                blury.compute_root()
                .reorder(c, x, y, z)
                .reorder_storage(c, x, y, z)
                .vectorize(c)
                .unroll(y, 2, hl.TailStrategy.RoundUp)
                .gpu_tile(x, y, z, xi, yi, zi, 32, 8, 1, hl.TailStrategy.RoundUp)
            )
            g.bilateral_grid.compute_root().gpu_tile(x, y, xi, yi, 32, 8)
            interpolated.compute_at(g.bilateral_grid, xi).vectorize(c)
        else:
            # CPU schedule.

            # 3.98ms on an Intel i9-9960X using 32 threads at 3.7 GHz
            # using target x86-64-avx2. This is a little less
            # SIMD-friendly than some of the other apps, so we
            # benefit from hyperthreading, and don't benefit from
            # AVX-512, which on my machine reduces the clock to 3.0
            # GHz.

            (
                blurz.compute_root()
                .reorder(c, z, x, y)
                .parallel(y)
                .vectorize(x, 8)
                .unroll(c)
            )
            histogram.compute_at(blurz, y)
            histogram.update().reorder(c, r.x, r.y, x, y).unroll(c)
            (
                blurx.compute_root()  #
                .reorder(c, x, y, z)  #
                .parallel(z)  #
                .vectorize(x, 8)  #
                .unroll(c)
            )
            (
                blury.compute_root()
                .reorder(c, x, y, z)
                .parallel(z)
                .vectorize(x, 8)
                .unroll(c)
            )
            g.bilateral_grid.compute_root().parallel(y).vectorize(x, 8)


if __name__ == "__main__":
    hl.main()



import halide as hl

x = hl.Var("x")
y = hl.Var("y")


@hl.generator(name="simplepy")
class SimplePy:
    offset = hl.GeneratorParam(0)

    buffer_input = hl.InputBuffer(hl.UInt(8), 2)

    # We can use `float` as an alias for `hl.Float(32)` if we like.
    float_arg = hl.InputScalar(float)

    simple_output = hl.OutputBuffer(hl.Float(32), 2)

    def generate(self):
        g = self

        g.simple_output[x, y] = hl.f32(g.offset + g.buffer_input[x, y]) + g.float_arg
        g.simple_output.compute_root()


if __name__ == "__main__":
    hl.main()



import halide as hl


def test_target():
    # Target("") should be exactly like get_host_target().
    t1 = hl.get_host_target()
    t2 = hl.Target("")
    assert t1 == t2, "Default ctor failure"
    assert t1.supported()

    # to_string roundtripping
    t1 = hl.Target()
    ts = t1.to_string()
    assert ts == "arch_unknown-0-os_unknown"

    # Note, this should *not* validate, since validate_target_string
    # now returns false if any of arch-bits-os are undefined
    assert not hl.Target.validate_target_string(ts)

    # Don't attempt to roundtrip this: trying to create
    # a Target with unknown portions will now assert-fail.
    #
    # t2 = hl.Target(ts)
    # assert t2 == t1

    # repr() and str()
    assert str(t1) == "arch_unknown-0-os_unknown"
    assert repr(t1) == "<halide.Target arch_unknown-0-os_unknown>"

    assert t1.os == hl.TargetOS.OSUnknown
    assert t1.arch == hl.TargetArch.ArchUnknown
    assert t1.bits == 0

    # Full specification round-trip:
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.SSE41])
    ts = t1.to_string()
    assert ts == "x86-32-linux-sse41"
    assert hl.Target.validate_target_string(ts)

    # Full specification (without features) round-trip:
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32)
    ts = t1.to_string()
    assert ts == "x86-32-linux"
    assert hl.Target.validate_target_string(ts)

    # Full specification round-trip, crazy features
    t1 = hl.Target(
        hl.TargetOS.Android,
        hl.TargetArch.ARM,
        32,
        [
            hl.TargetFeature.JIT,
            hl.TargetFeature.CUDA,
            hl.TargetFeature.OpenCL,
            hl.TargetFeature.OpenGLCompute,
            hl.TargetFeature.Debug,
        ],
    )
    ts = t1.to_string()
    assert ts == "arm-32-android-cuda-debug-jit-opencl-openglcompute"
    assert hl.Target.validate_target_string(ts)

    # Expected failures:
    ts = "host-unknowntoken"
    assert not hl.Target.validate_target_string(ts)

    ts = "x86-23"
    assert not hl.Target.validate_target_string(ts)

    # bits == 0 is allowed only if arch_unknown and os_unknown are specified,
    # and no features are set
    ts = "x86-0"
    assert not hl.Target.validate_target_string(ts)

    ts = "0-arch_unknown-os_unknown-sse41"
    assert not hl.Target.validate_target_string(ts)

    # "host" is only supported as the first token
    ts = "opencl-host"
    assert not hl.Target.validate_target_string(ts)

    # set_feature
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.SSE41])
    assert t1.has_feature(hl.TargetFeature.SSE41)
    assert not t1.has_feature(hl.TargetFeature.AVX)
    t1.set_feature(hl.TargetFeature.AVX)
    t1.set_feature(hl.TargetFeature.SSE41, False)
    assert t1.has_feature(hl.TargetFeature.AVX)
    assert not t1.has_feature(hl.TargetFeature.SSE41)

    # set_features
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.SSE41])
    assert t1.has_feature(hl.TargetFeature.SSE41)
    assert not t1.has_feature(hl.TargetFeature.AVX)
    t1.set_features([hl.TargetFeature.SSE41], False)
    t1.set_features([hl.TargetFeature.AVX, hl.TargetFeature.AVX2], True)
    assert t1.has_feature(hl.TargetFeature.AVX)
    assert t1.has_feature(hl.TargetFeature.AVX2)
    assert not t1.has_feature(hl.TargetFeature.SSE41)

    # with_feature
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.SSE41])
    t2 = t1.with_feature(hl.TargetFeature.NoAsserts).with_feature(
        hl.TargetFeature.NoBoundsQuery
    )
    ts = t2.to_string()
    assert ts == "x86-32-linux-no_asserts-no_bounds_query-sse41"

    # without_feature
    t1 = hl.Target(
        hl.TargetOS.Linux,
        hl.TargetArch.X86,
        32,
        [hl.TargetFeature.SSE41, hl.TargetFeature.NoAsserts],
    )
    # Note that NoBoundsQuery wasn't set here, so 'without' is a no-op
    t2 = t1.without_feature(hl.TargetFeature.NoAsserts).without_feature(
        hl.TargetFeature.NoBoundsQuery
    )
    ts = t2.to_string()
    assert ts == "x86-32-linux-sse41"

    # natural_vector_size
    # SSE4.1 is 16 bytes wide
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.SSE41])
    assert t1.natural_vector_size(hl.UInt(8)) == 16
    assert t1.natural_vector_size(hl.Int(16)) == 8
    assert t1.natural_vector_size(hl.UInt(32)) == 4
    assert t1.natural_vector_size(hl.Float(32)) == 4

    # has_gpu_feature
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [hl.TargetFeature.OpenCL])
    t2 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, [])
    assert t1.has_gpu_feature()
    assert not t2.has_gpu_feature()

    # has_large_buffers & maximum_buffer_size
    t1 = hl.Target(
        hl.TargetOS.Linux,
        hl.TargetArch.X86,
        64,
        [hl.TargetFeature.LargeBuffers],
    )
    t2 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 64, [])
    assert t1.has_large_buffers()
    assert t1.maximum_buffer_size() == 9223372036854775807
    assert not t2.has_large_buffers()
    assert t2.maximum_buffer_size() == 2147483647

    # supports_device_api
    t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 64, [hl.TargetFeature.CUDA])
    t2 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 64)
    assert t1.supports_device_api(hl.DeviceAPI.CUDA)
    assert not t2.supports_device_api(hl.DeviceAPI.CUDA)

    # supports_type (deprecated version)
    t1 = hl.Target(hl.TargetOS.OSX, hl.TargetArch.X86, 64, [hl.TargetFeature.Metal])
    t2 = hl.Target(hl.TargetOS.OSX, hl.TargetArch.X86, 64)
    assert not t1.supports_type(hl.Float(64))
    assert t2.supports_type(hl.Float(64))

    # supports_type (preferred version)
    t1 = hl.Target(hl.TargetOS.OSX, hl.TargetArch.X86, 64, [hl.TargetFeature.Metal])
    t2 = hl.Target(hl.TargetOS.OSX, hl.TargetArch.X86, 64)
    assert not t1.supports_type(hl.Float(64), hl.DeviceAPI.Metal)
    assert not t2.supports_type(hl.Float(64), hl.DeviceAPI.Metal)

    # target_feature_for_device_api
    assert (
        hl.target_feature_for_device_api(hl.DeviceAPI.OpenCL) == hl.TargetFeature.OpenCL
    )

    # with_feature with non-convertible lists
    try:
        t1 = hl.Target(hl.TargetOS.Linux, hl.TargetArch.X86, 32, ["this is a string"])
    except TypeError as e:
        assert "incompatible constructor arguments" in str(e)
    else:
        assert False, "Did not see expected exception!"


if __name__ == "__main__":
    test_target()



import numpy as np

from multi_method_module import simplecpp, user_context


def test_simplecpp():
    buffer_input = np.ndarray([2, 2], dtype=np.uint8)
    buffer_input[0, 0] = 123
    buffer_input[0, 1] = 123
    buffer_input[1, 0] = 123
    buffer_input[1, 1] = 123

    float_arg = 3.5

    simple_output = np.ndarray([2, 2], dtype=np.float32)

    simplecpp(buffer_input, float_arg, simple_output)

    assert simple_output[0, 0] == 3.5 + 123
    assert simple_output[0, 1] == 3.5 + 123
    assert simple_output[1, 0] == 3.5 + 123
    assert simple_output[1, 1] == 3.5 + 123


def test_user_context():
    output = bytearray("\0\0\0\0", "ascii")
    user_context(None, ord("q"), output)
    assert output == bytearray("qqqq", "ascii")


def test_aot_call_failure_throws_exception():
    buffer_input = np.zeros([2, 2], dtype=np.float32)  # wrong type
    float_arg = 3.5
    simple_output = np.zeros([2, 2], dtype=np.float32)

    try:
        simplecpp(buffer_input, float_arg, simple_output)
    except RuntimeError as e:
        assert "Halide Runtime Error: -3" in str(e), str(e)
    else:
        assert False, "Did not see expected exception, saw: " + str(e)


if __name__ == "__main__":
    test_simplecpp()
    test_user_context()
    test_aot_call_failure_throws_exception()



import math
import sys
import warnings

import halide as hl
import numpy


class AssertWarnsContext:
    def __init__(self, warn):
        self.expected = warn
        self.occurred = False

    def __bool__(self):
        return self.occurred

    def __enter__(self):
        self.warnings_manager = warnings.catch_warnings(record=True)
        self.warnings = self.warnings_manager.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.warnings_manager.__exit__(exc_type, exc_value, tb)
        if exc_type is not None:
            return
        self.occurred = False
        for m in self.warnings:
            if not isinstance(m.message, self.expected):
                continue
            self.occurred = True


def test():
    def test_pattern(c):
        x = hl.Var("x")
        f = hl.Func("f")
        f[x] = x * hl.f64(c) * (hl.f64(0.1) + hl.f64(0.2))
        for i, hl_value in enumerate(numpy.asarray(f.realize([10]))):
            py_value = i * c * (0.1 + 0.2)
            check = math.isclose(hl_value, py_value)
            assert check, "{}[{}]: {} != {}".format(i, c, hl_value, py_value)

    test_pattern(0.123456789012345678)
    test_pattern(0.987654321098765432)

    x = hl.Var("x")
    with AssertWarnsContext(RuntimeWarning) as ctx:
        x + 0.123456789012345678
    assert ctx.occurred, "RuntimeWarning didn't occur."

    with AssertWarnsContext(RuntimeWarning) as ctx:
        x + hl.f64(0.123456789012345678)
    assert not ctx.occurred, "RuntimeWarning occurred."

    with AssertWarnsContext(RuntimeWarning) as ctx:
        x + 0.75  # 0.5 + 0.25
    assert not ctx.occurred, "RuntimeWarning occurred."


if __name__ == "__main__":
    test()



"""
Simple blur.
"""

from blur import blur
import halide.imageio
import numpy as np
import sys
import timeit


def main():
    if len(sys.argv) < 3:
        print("Usage: %s input.png output.png" % sys.argv[0])
        print("e.g. %s input.png output.png 10" % sys.argv[0])
        sys.exit(1)

    input_path = sys.argv[1]
    output_path = sys.argv[2]
    timing_iterations = 10

    print("Reading from %s ..." % input_path)
    input_buf_u8 = halide.imageio.imread(input_path)
    assert input_buf_u8.dtype == np.uint8
    # Convert to uint16... but remember that the blur() generator
    # is documented as only working on <= 14 bits of image; if
    # we use the upper two bits we'll get incorrect results.
    # We'll just leave it with 8 bits of useful data.
    input_buf = input_buf_u8.astype(np.uint16)
    output_buf = np.empty(input_buf.shape, dtype=input_buf.dtype)

    tests = {
        "Manual": blur,
    }

    for name, fn in tests.items():
        print("Running %s... " % name, end="")
        t = timeit.Timer(lambda: fn(input_buf, output_buf))
        avg_time_sec = t.timeit(number=timing_iterations) / timing_iterations
        print("time: %fms" % (avg_time_sec * 1e3))

    output_buf_u8 = output_buf.astype(np.uint8)
    print("Saving to %s ..." % output_path)
    halide.imageio.imwrite(output_path, output_buf_u8)

    print("Success!")
    sys.exit(0)


if __name__ == "__main__":
    main()
